# -*- coding: utf-8 -*-


try:
    from collections.abc import Mapping
except ImportError:
    from collections import Mapping

from functools import reduce
from itertools import product
from operator import mul

import numpy as np

from africanus.util.requirements import requires_optional

from africanus.rime.predict import (PREDICT_DOCS, predict_checks,
                                    predict_vis as np_predict_vis)

try:
    from dask.blockwise import blockwise
    from dask.base import tokenize
    import dask.array as da
    from dask.highlevelgraph import HighLevelGraph
except ImportError as e:
    opt_import_error = e
else:
    opt_import_error = None


def _source_stream_blocks(source_blocks, streams):
    return (source_blocks + streams - 1) // streams


def _extract_blocks(time_index, dde1_jones, source_coh, dde2_jones):
    """
    Returns
    -------
    blocks : tuple
        :code:`(source, row, ant, chan, corr1, ..., corrn)
    """

    if dde1_jones is not None:
        return ((dde1_jones.numblocks[0], time_index.numblocks[0]) +
                (1, dde1_jones.numblocks[3]) +
                dde1_jones.numblocks[4:])
    elif source_coh is not None:
        return (source_coh.numblocks[:2] +
                (1, source_coh.numblocks[2]) +
                source_coh.numblocks[3:])
    else:
        raise ValueError("need ddes or coherencies")


def _extract_chunks(time_index, dde1_jones, source_coh, dde2_jones):
    """
    Returns
    -------
    chunks : tuple
        :code:`(source, row, chan, corr1, ..., corrn)
    """

    if dde1_jones is not None:
        return ((dde1_jones.chunks[0], time_index.chunks[0]) +
                (dde1_jones.chunks[3],) +
                dde1_jones.chunks[4:])
    elif source_coh is not None:
        return source_coh.chunks
    else:
        raise ValueError("need ddes or coherencies")


class CoherencyStreamReduction(Mapping):
    """
    tl;dr this is a dictionary that is expanded in place when
    first acccessed. Saves memory when pickled for sending
    to the dask scheduler.

    See :class:`dask.blockwise.Blockwise` for further insight.

    Produces graph serially summing coherencies in
    ``stream`` parallel streams.
    """

    def __init__(self, time_index, antenna1, antenna2,
                 dde1_jones, source_coh, dde2_jones,
                 out_name, streams):
        self.time_index_name = None if time_index is None else time_index.name
        self.ant1_name = None if antenna1 is None else antenna1.name
        self.ant2_name = None if antenna2 is None else antenna2.name
        self.dde1_name = None if dde1_jones is None else dde1_jones.name
        self.coh_name = None if source_coh is None else source_coh.name
        self.dde2_name = None if dde2_jones is None else dde2_jones.name

        self.out_name = out_name

        self.blocks = _extract_blocks(time_index, dde1_jones,
                                      source_coh, dde2_jones)
        self.streams = streams

    @property
    def _dict(self):
        if hasattr(self, "_cached_dict"):
            return self._cached_dict
        else:
            self._cached_dict = self._create_dict()
            return self._cached_dict

    def __getitem__(self, key):
        return self._dict[key]

    def __iter__(self):
        return iter(self._dict)

    def __len__(self):
        # Extract dimension blocks
        (source, row, _, chan), corr = self.blocks[:4], self.blocks[4:]
        return reduce(mul, (source, row, chan) + corr, 1)

    def _create_dict(self):
        # Graph dictionary
        layers = {}

        # For loop performance
        out_name = self.out_name
        ti = self.time_index_name
        a1 = self.ant1_name
        a2 = self.ant2_name
        dde1 = self.dde1_name
        coh = self.coh_name
        dde2 = self.dde2_name

        # Extract dimension blocks
        (source_blocks, row_blocks, ant_blocks,
         chan_blocks), corr_blocks = self.blocks[:4], self.blocks[4:]

        assert ant_blocks == 1
        ab = 0

        # Subdivide number of source blocks by number of streams
        source_block_chunks = _source_stream_blocks(source_blocks,
                                                    self.streams)

        # Iterator of block id's for row, channel and correlation blocks
        # We don't reduce over these dimensions
        block_ids = enumerate(product(range(row_blocks), range(chan_blocks),
                                      *[range(cb) for cb in corr_blocks]))

        for flat_bid, bid in block_ids:
            rb, fb = bid[0:2]
            cb = bid[2:]

            # Create the streamed reduction proper.
            # For a stream, the base visibilities are set to the result
            # of the previous result in the stream (last_key)
            for sb_start in range(0, source_blocks, source_block_chunks):
                sb_end = min(sb_start + source_block_chunks, source_blocks)
                last_key = None

                for sb in range(sb_start, sb_end):
                    # Dask task object calling predict vis
                    task = (np_predict_vis,
                            (ti, rb), (a1, rb), (a2, rb),
                            (dde1, sb, rb, ab, fb) + cb if dde1 else None,
                            (coh, sb, rb, fb) + cb if coh else None,
                            (dde2, sb, rb, ab, fb) + cb if dde2 else None,
                            None, last_key, None)

                    key = (out_name, sb, flat_bid)
                    layers[key] = task
                    last_key = key

        return layers


class CoherencyFinalReduction(Mapping):
    """
    tl;dr this is a dictionary that is expanded in place when
    first acccessed. Saves memory when pickled for sending
    to the dask scheduler.

    See :class:`dask.blockwise.Blockwise` for further insight.

    Produces graph reducing results of ``stream`` parallel streams in
    CoherencyStreamReduction.
    """

    def __init__(self, out_name, coherency_stream_reduction):
        self.in_name = coherency_stream_reduction.out_name
        self.blocks = coherency_stream_reduction.blocks
        self.streams = coherency_stream_reduction.streams
        self.out_name = out_name

    @property
    def _dict(self):
        if hasattr(self, "_cached_dict"):
            return self._cached_dict
        else:
            self._cached_dict = self._create_dict()
            return self._cached_dict

    def __getitem__(self, key):
        return self._dict[key]

    def __iter__(self):
        return iter(self._dict)

    def __len__(self):
        (source, row, _, chan), corrs = self.blocks[:4], self.blocks[4:]
        return reduce(mul, (source, row, chan) + corrs, 1)

    def _create_dict(self):
        (source, row, _, chan), corrs = self.blocks[:4], self.blocks[4:]

        # Iterator of block id's for row, channel and correlation blocks
        # We don't reduce over these dimensions
        block_ids = enumerate(product(range(row), range(chan),
                                      *[range(cb) for cb in corrs]))

        source_block_chunks = _source_stream_blocks(source, self.streams)

        layers = {}

        # This looping structure should match
        for flat_bid, bid in block_ids:
            rb, fb = bid[0:2]
            cb = bid[2:]

            last_stream_keys = []

            for sb_start in range(0, source, source_block_chunks):
                sb_end = min(sb_start + source_block_chunks, source)
                key = (sb_end - 1, flat_bid)
                last_stream_keys.append((self.in_name, sb_end - 1, flat_bid))

            key = (self.out_name, rb, fb) + cb
            task = (sum, last_stream_keys)
            layers[key] = task

        return layers


def _predict_coh_wrapper(time_index, antenna1, antenna2,
                         dde1_jones, source_coh, dde2_jones,
                         die1_jones, base_vis, die2_jones):

    return (np_predict_vis(time_index, antenna1, antenna2,
                           # dde1_jones loses the 'ant' dim
                           dde1_jones[0] if dde1_jones else None,
                           # source_coh loses the 'source' dim
                           source_coh,
                           # dde2_jones loses the 'source' and 'ant' dims
                           dde2_jones[0] if dde2_jones else None,
                           # die1_jones loses the 'ant' dim
                           die1_jones[0] if die1_jones else None,
                           base_vis,
                           # die2_jones loses the 'ant' dim
                           die2_jones[0] if die2_jones else None)
            # Introduce an extra dimension (source dim reduced to 1)
            [None, ...])


def _predict_dies_wrapper(time_index, antenna1, antenna2,
                          dde1_jones, source_coh, dde2_jones,
                          die1_jones, base_vis, die2_jones):

    return np_predict_vis(time_index, antenna1, antenna2,
                          # dde1_jones loses the 'source' and 'ant' dims
                          dde1_jones[0][0] if dde1_jones else None,
                          # source_coh loses the 'source' dim
                          source_coh[0] if source_coh else None,
                          # dde2_jones loses the 'source' and 'ant' dims
                          dde2_jones[0][0] if dde2_jones else None,
                          # die1_jones loses the 'ant' dim
                          die1_jones[0] if die1_jones else None,
                          base_vis,
                          # die2_jones loses the 'ant' dim
                          die2_jones[0] if die2_jones else None)


def stream_reduction(time_index, antenna1, antenna2,
                     dde1_jones, source_coh, dde2_jones,
                     predict_check_tup, out_dtype, streams):
    """
    Reduces source coherencies + ddes over the source dimension in
    ``N`` parallel streams.

    This is accomplished by calling predict_vis on on ddes and source
    coherencies to produce visibilities which are passed into
    the `base_vis` argument of ``predict_vis`` for the next chunk.
    """

    # Unique name and token for this operation
    token = tokenize(time_index, antenna1, antenna2,
                     dde1_jones, source_coh, dde2_jones,
                     streams)

    name = 'stream-coherency-reduction-' + token

    # Number of dim blocks
    blocks = _extract_blocks(time_index, dde1_jones, source_coh, dde2_jones)
    (src_blocks, row_blocks, _,
     chan_blocks), corr_blocks = blocks[:4], blocks[4:]

    # Total number of other dimension blocks
    nblocks = reduce(mul, (row_blocks, chan_blocks) + corr_blocks, 1)

    # Create the compressed mapping
    layers = CoherencyStreamReduction(time_index, antenna1, antenna2,
                                      dde1_jones, source_coh, dde2_jones,
                                      name, streams)

    # Create the graph
    extra_deps = [a for a in (dde1_jones, source_coh, dde2_jones)
                  if a is not None]
    deps = [time_index, antenna1, antenna2] + extra_deps

    graph = HighLevelGraph.from_collections(name, layers, deps)

    chunks = ((1,) * src_blocks, (1,)*nblocks)
    # This should never be directly computed, reported chunks
    # and dtype don't match the actual data. We create it
    # because it makes chaining HighLevelGraphs easier
    stream_reduction = da.Array(graph, name, chunks, dtype=np.int8)

    name = "coherency-reduction-" + tokenize(stream_reduction)
    layers = CoherencyFinalReduction(name, layers)
    graph = HighLevelGraph.from_collections(name, layers, [stream_reduction])

    chunks = _extract_chunks(time_index, dde1_jones, source_coh, dde2_jones)
    return da.Array(graph, name, chunks[1:], dtype=out_dtype)


def fan_reduction(time_index, antenna1, antenna2,
                  dde1_jones, source_coh, dde2_jones,
                  predict_check_tup, out_dtype):
    """ Does a standard dask tree reduction over source coherencies """
    (have_ddes1, have_coh, have_ddes2,
     have_dies1, have_bvis, have_dies2) = predict_check_tup

    have_ddes = have_ddes1 and have_ddes2

    if have_ddes:
        cdims = tuple("corr-%d" % i for i in range(len(dde1_jones.shape[4:])))
    elif have_coh:
        cdims = tuple("corr-%d" % i for i in range(len(source_coh.shape[3:])))
    else:
        raise ValueError("need ddes or source coherencies")

    ajones_dims = ("src", "row", "ant", "chan") + cdims

    # Setup
    # 1. Optional blockwise arguments
    # 2. Optional numblocks kwarg
    # 3. HighLevelGraph dependencies
    bw_args = [time_index.name, ("row",),
               antenna1.name, ("row",),
               antenna2.name, ("row",)]
    numblocks = {
        time_index.name: time_index.numblocks,
        antenna1.name: antenna1.numblocks,
        antenna2.name: antenna2.numblocks
    }

    # Dependencies
    deps = [time_index, antenna1, antenna2]

    # Handle presence/absence of dde1_jones
    if have_ddes:
        bw_args.extend([dde1_jones.name, ajones_dims])
        numblocks[dde1_jones.name] = dde1_jones.numblocks
        deps.append(dde1_jones)
        other_chunks = dde1_jones.chunks[3:]
        src_chunks = dde1_jones.chunks[0]
    else:
        bw_args.extend([None, None])

    # Handle presence/absence of source_coh
    if have_coh:
        bw_args.extend([source_coh.name, ("src", "row", "chan") + cdims])
        numblocks[source_coh.name] = source_coh.numblocks
        deps.append(source_coh)
        other_chunks = source_coh.chunks[2:]
        src_chunks = source_coh.chunks[0]
    else:
        bw_args.extend([None, None])

    # Handle presence/absence of dde2_jones
    if have_ddes:
        bw_args.extend([dde2_jones.name, ajones_dims])
        numblocks[dde2_jones.name] = dde2_jones.numblocks
        deps.append(dde2_jones)
        other_chunks = dde2_jones.chunks[3:]
        src_chunks = dde2_jones.chunks[0]
    else:
        bw_args.extend([None, None])

    # die1_jones, base_vis and die2_jones absent for this part of the graph
    bw_args.extend([None, None, None, None, None, None])

    assert len(bw_args) // 2 == 9, len(bw_args) // 2

    token = da.core.tokenize(time_index, antenna1, antenna2,
                             dde1_jones, source_coh, dde2_jones)
    name = "-".join(("predict-vis-sum-coh", token))
    layer = blockwise(_predict_coh_wrapper,
                      name, ("src", "row", "chan") + cdims,
                      *bw_args, numblocks=numblocks)

    graph = HighLevelGraph.from_collections(name, layer, deps)

    # We can infer output chunk sizes from source_coh
    chunks = ((1,)*len(src_chunks), time_index.chunks[0],) + other_chunks

    # Create array
    sum_coherencies = da.Array(graph, name, chunks, dtype=out_dtype)

    # Reduce source axis
    return sum_coherencies.sum(axis=0)


def apply_dies(time_index, antenna1, antenna2,
               die1_jones, base_vis, die2_jones,
               predict_check_tup, out_dtype):
    """ Apply any Direction-Independent Effects and Base Visibilities """

    # Now apply any Direction Independent Effect Terms
    (have_ddes1, have_coh, have_ddes2,
     have_dies1, have_bvis, have_dies2) = predict_check_tup

    have_dies = have_dies1 and have_dies2

    # Generate strings for the correlation dimensions
    # This also has the effect of checking that we have all valid inputs
    if have_dies:
        cdims = tuple("corr-%d" % i for i in range(len(die1_jones.shape[3:])))
    elif have_bvis:
        cdims = tuple("corr-%d" % i for i in range(len(base_vis.shape[2:])))
    else:
        raise ValueError("Missing both antenna and baseline jones terms")

    # In the case of predict_vis, the "row" and "time" dimensions
    # are intimately related -- a contiguous series of rows
    # are related to a contiguous series of timesteps.
    # This means that the number of chunks of these
    # two dimensions must match even though the chunk sizes may not.
    # blockwise insists on matching chunk sizes.
    # For this reason, we use the lower level blockwise and
    # substitute "row" for "time" in arrays such as dde1_jones
    # and die1_jones.
    gjones_dims = ("row", "ant", "chan") + cdims

    # Setup
    # 1. Optional blockwise arguments
    # 2. Optional numblocks kwarg
    # 3. HighLevelGraph dependencies
    bw_args = [time_index.name, ("row",),
               antenna1.name, ("row",),
               antenna2.name, ("row",)]
    numblocks = {
        time_index.name: time_index.numblocks,
        antenna1.name: antenna1.numblocks,
        antenna2.name: antenna2.numblocks
    }

    deps = [time_index, antenna1, antenna2]

    # dde1_jones, source_coh and dde2_jones not present
    # these are already applied into sum_coherencies
    bw_args.extend([None, None, None, None, None, None])

    if have_dies:
        bw_args.extend([die1_jones.name, gjones_dims])
        numblocks[die1_jones.name] = die1_jones.numblocks
        deps.append(die1_jones)
        other_chunks = die1_jones.chunks[2:]
    else:
        bw_args.extend([None, None])

    if have_bvis:
        bw_args.extend([base_vis.name, ("row", "chan") + cdims])
        numblocks[base_vis.name] = base_vis.numblocks
        deps.append(base_vis)
        other_chunks = base_vis.chunks[1:]
    else:
        bw_args.extend([None, None])

    if have_dies:
        bw_args.extend([die2_jones.name, gjones_dims])
        numblocks[die2_jones.name] = die2_jones.numblocks
        deps.append(die2_jones)
        other_chunks = die2_jones.chunks[2:]
    else:
        bw_args.extend([None, None])

    assert len(bw_args) // 2 == 9

    token = da.core.tokenize(time_index, antenna1, antenna2,
                             die1_jones, base_vis, die2_jones)
    name = '-'.join(("predict-vis-apply-dies", token))
    layer = blockwise(_predict_dies_wrapper,
                      name, ("row", "chan") + cdims,
                      *bw_args, numblocks=numblocks)

    graph = HighLevelGraph.from_collections(name, layer, deps)
    chunks = (time_index.chunks[0],) + other_chunks

    return da.Array(graph, name, chunks, dtype=out_dtype)


@requires_optional('dask.array', opt_import_error)
def predict_vis(time_index, antenna1, antenna2,
                dde1_jones=None, source_coh=None, dde2_jones=None,
                die1_jones=None, base_vis=None, die2_jones=None,
                streams=None):

    predict_check_tup = predict_checks(time_index, antenna1, antenna2,
                                       dde1_jones, source_coh, dde2_jones,
                                       die1_jones, base_vis, die2_jones)

    (have_ddes1, have_coh, have_ddes2,
     have_dies1, have_bvis, have_dies2) = predict_check_tup

    have_ddes = have_ddes1 and have_ddes2

    if have_ddes:
        if dde1_jones.shape[2] != dde1_jones.chunks[2][0]:
            raise ValueError("Subdivision of antenna dimension into "
                             "multiple chunks is not supported.")

        if dde2_jones.shape[2] != dde2_jones.chunks[2][0]:
            raise ValueError("Subdivision of antenna dimension into "
                             "multiple chunks is not supported.")

        if dde1_jones.chunks != dde2_jones.chunks:
            raise ValueError("dde1_jones.chunks != dde2_jones.chunks")

        if len(dde1_jones.chunks[1]) != len(time_index.chunks[0]):
            raise ValueError("Number of row chunks (%s) does not equal "
                             "number of time chunks (%s)." %
                             (time_index.chunks[0], dde1_jones.chunks[1]))

    have_dies = have_dies1 and have_dies2

    if have_dies:
        if die1_jones.shape[1] != die1_jones.chunks[1][0]:
            raise ValueError("Subdivision of antenna dimension into "
                             "multiple chunks is not supported.")

        if die2_jones.shape[1] != die2_jones.chunks[1][0]:
            raise ValueError("Subdivision of antenna dimension into "
                             "multiple chunks is not supported.")

        if die1_jones.chunks != die2_jones.chunks:
            raise ValueError("die1_jones.chunks != die2_jones.chunks")

        if len(die1_jones.chunks[0]) != len(time_index.chunks[0]):
            raise ValueError("Number of row chunks (%s) does not equal "
                             "number of time chunks (%s)." %
                             (time_index.chunks[0], die1_jones.chunks[1]))

    # Infer the output dtype
    dtype_arrays = [dde1_jones, source_coh, dde2_jones, die1_jones, die2_jones]
    out_dtype = np.result_type(*(np.dtype(a.dtype.name)
                                 for a in dtype_arrays if a is not None))

    # Apply direction dependent effects
    if have_coh or have_ddes:
        # We create separate graphs for computing coherencies and applying
        # the gains because coherencies are chunked over source which
        # must be summed and added to the (possibly present) base visibilities
        if streams is not None:
            sum_coherencies = stream_reduction(time_index, antenna1, antenna2,
                                               dde1_jones, source_coh,
                                               dde2_jones, predict_check_tup,
                                               out_dtype, streams=streams)
        else:
            sum_coherencies = fan_reduction(time_index, antenna1, antenna2,
                                            dde1_jones, source_coh, dde2_jones,
                                            predict_check_tup, out_dtype)
    else:
        assert have_dies or have_bvis
        sum_coherencies = None

    # No more effects to apply, return at this point
    if not have_dies and not have_bvis:
        return sum_coherencies

    # Add coherencies to the base visibilities
    if sum_coherencies is not None:
        if not have_bvis:
            # Set base_vis = summed coherencies
            base_vis = sum_coherencies
            predict_check_tup = (have_ddes1, have_coh, have_ddes2,
                                 have_dies1, True, have_dies2)
        else:
            base_vis += sum_coherencies

    # Apply direction independent effects
    return apply_dies(time_index, antenna1, antenna2,
                      die1_jones, base_vis, die2_jones,
                      predict_check_tup, out_dtype)


EXTRA_DASK_ARGS = """
streams : int, optional
    Specifies the degree of parallelism along the source dimension.
    By default, dask uses a tree style reduction algorithm which can
    require large amounts of memory. Specifying this parameter
    constrains the dask graph to serially sum coherencies in a
    specified number of streams, reducing overall memory usage.

    If ``None``, defaults to a standard, memory-intensive tree style
    algorithm.

    Defaults to 1, which means that the source coherencies for each
    visibility chunk are serially summed, meaning that parallelism
    will only exists along the row and chan dimensions.
"""

EXTRA_DASK_NOTES = """
* The ``ant`` dimension should only contain a single chunk equal
  to the number of antenna. Since each ``row`` can contain
  any antenna, random access must be preserved along this dimension.
* The chunks in the ``row`` and ``time`` dimension **must** align.
  This subtle point **must be understood otherwise
  invalid results will be produced** by the chunking scheme.
  In the example below
  we have four unique time indices :code:`[0,1,2,3]`, and
  four unique antenna :code:`[0,1,2,3]` indexing :code:`10` rows.

  .. code-block:: python

      #  Row indices into the time/antenna indexed arrays
      time_idx = np.asarray([0,0,1,1,2,2,2,2,3,3])
      ant1 = np.asarray(    [0,0,0,0,1,1,1,2,2,3]
      ant2 = np.asarray(    [0,1,2,3,1,2,3,2,3,3])


  A reasonable chunking scheme for the
  ``row`` and ``time`` dimension would be :code:`(4,4,2)`
  and :code:`(2,1,1)` respectively.
  Another way of explaining this is that the first
  four rows contain two unique timesteps, the second four
  rows contain one unique timestep and the last two rows
  contain one unique timestep.

  Some rules of thumb:

  1. The number chunks in ``row`` and ``time`` must match
     although the individual chunk sizes need not.
  2. Unique timesteps should not be split across row chunks.
  3. For a Measurement Set whose rows are ordered on the
     ``TIME`` column, the following is a good way of obtaining
     the row chunking strategy:

     .. code-block:: python

        import numpy as np
        import pyrap.tables as pt

        ms = pt.table("data.ms")
        times = ms.getcol("TIME")
        unique_times, chunks = np.unique(times, return_counts=True)

  4. Use :func:`~africanus.util.shapes.aggregate_chunks`
     to aggregate multiple ``row`` and ``time``
     chunks into chunks large enough such that functions operating
     on the resulting data can drop the GIL and spend time
     processing the data. Expanding the previous example:

     .. code-block:: python

        # Aggregate row
        utimes = unique_times.size
        # Single chunk for each unique time
        time_chunks = (1,)*utimes
        # Aggregate row chunks into chunks <= 10000
        aggregate_chunks((chunks, time_chunks), (10000, utimes))
"""

try:
    predict_vis.__doc__ = PREDICT_DOCS.substitute(
                                array_type=":class:`dask.array.Array`",
                                get_time_index=":code:`time.map_blocks("
                                               "lambda a: np.unique(a, "
                                               "return_inverse=True)[1])`",
                                extra_args=EXTRA_DASK_ARGS,
                                extra_notes=EXTRA_DASK_NOTES)
except AttributeError:
    pass



# -*- coding: utf-8 -*-


import numpy as np
from africanus.calibration.utils import check_type
from africanus.util.docs import DocstringTemplate
from africanus.calibration.utils import residual_vis
from africanus.util.numba import generated_jit, njit

DIAG_DIAG = 0
DIAG = 1
FULL = 2


def jacobian_factory(mode):
    if mode == DIAG_DIAG:
        def jacobian(a1j, blj, a2j, sign, out):
            out[...] = sign * a1j * blj * np.conj(a2j)
    elif mode == DIAG:
        def jacobian(a1j, blj, a2j, sign, out):
            out[...] = 0
    elif mode == FULL:
        def jacobian(a1j, blj, a2j, sign, out):
            out[...] = 0
    return njit(nogil=True)(jacobian)


@generated_jit(nopython=True, nogil=True, cache=True)
def jhj_and_jhr(time_bin_indices, time_bin_counts, antenna1,
                antenna2, jones, residual, model, flag):

    mode = check_type(jones, residual)

    if mode:
        raise NotImplementedError("Only DIAG-DIAG case has been implemented")

    jacobian = jacobian_factory(mode)

    def _jhj_and_jhr_fn(time_bin_indices, time_bin_counts, antenna1,
                        antenna2, jones, residual, model, flag):
        jones_shape = np.shape(jones)
        tmp_out_array = np.zeros_like(jones[0, 0, 0, 0], dtype=jones.dtype)
        n_tim = jones_shape[0]
        n_ant = jones_shape[1]
        n_chan = jones_shape[2]
        n_dir = jones_shape[3]

        jhr = np.zeros(jones.shape, dtype=jones.dtype)
        jhj = np.zeros(jones.shape, dtype=jones.real.dtype)

        for t in range(n_tim):
            ind = np.arange(time_bin_indices[t],
                            time_bin_indices[t] + time_bin_counts[t])
            for ant in range(n_ant):
                # find where either antenna == ant
                # these will be mutually exclusive since no autocorrs
                for row in ind:
                    if antenna1[row] == ant or antenna2[row] == ant:
                        p = antenna1[row]
                        q = antenna2[row]
                        if ant == p:
                            sign = 1.0j
                        elif ant == q:
                            sign = -1.0j
                        else:
                            raise ValueError(
                                "Got impossible antenna number. This is a bug")
                        for nu in range(n_chan):
                            if not np.any(flag[row, nu]):
                                for s in range(n_dir):
                                    jacobian(
                                        jones[t, p, nu, s],
                                        model[row, nu, s],
                                        jones[t, q, nu, s],
                                        sign,
                                        tmp_out_array)
                                    jhj[t, ant, nu,
                                        s] += (np.conj(tmp_out_array) *
                                               tmp_out_array).real
                                    jhr[t, ant, nu,
                                        s] += (np.conj(tmp_out_array) *
                                               residual[row, nu])
        return jhj, jhr
    return _jhj_and_jhr_fn


def phase_only_gauss_newton(time_bin_indices, time_bin_counts, antenna1,
                            antenna2, jones, vis, flag, model,
                            weight, tol=1e-4, maxiter=100):
    # whiten data
    sqrtweights = np.sqrt(weight)
    vis *= sqrtweights
    model *= sqrtweights[:, :, None]

    eps = 1.0
    k = 0
    while eps > tol and k < maxiter:
        # keep track of old phases
        phases = np.angle(jones)

        # get residual
        residual = residual_vis(time_bin_indices, time_bin_counts, antenna1,
                                antenna2, jones, vis, flag, model)

        # get diag(jhj) and jhr
        jhj, jhr = jhj_and_jhr(time_bin_indices, time_bin_counts, antenna1,
                               antenna2, jones, residual, model, flag)

        # implement update
        phases_new = phases + (jhr/jhj).real
        jones = np.exp(1.0j * phases_new)

        # check convergence/iteration control
        eps = np.abs(phases_new - phases).max()
        k += 1

    return jones, jhj, jhr, k


PHASE_CALIBRATION_DOCS = DocstringTemplate("""
Performs phase-only maximum likelihood
calibration assuming scalar or diagonal
inputs using Gauss-Newton oprimisation.

Parameters
----------
time_bin_indices : $(array_type)
    The start indices of the time bins
    of shape :code:`(utime)`
time_bin_counts : $(array_type)
    The counts of unique time in each
    time bin of shape :code:`(utime)`
antenna1 : $(array_type)
    First antenna indices of shape :code:`(row,)`.
antenna2 : $(array_type)
    Second antenna indices of shape :code:`(row,)`.
jones : $(array_type)
    Gain solutions of shape :code:`(time, ant, chan, dir, corr)`
    or :code:`(time, ant, chan, dir, corr, corr)`.
vis : $(array_type)
    Data values of shape :code:`(row, chan, corr)`
    or :code:`(row, chan, corr, corr)`.
flag : $(array_type)
    Flag data of shape :code:`(row, chan, corr)`
    or :code:`(row, chan, corr, corr)`.
model : $(array_type)
    Model data values of shape :code:`(row, chan, dir, corr)`
    or :code:`(row, chan, dir, corr, corr)`.
weight : $(array_type)
    Weight spectrum of shape :code:`(row, chan, corr)`.
    If the channel axis is missing weights are duplicated
    for each channel.
tol : float, optional
    The tolerance of the solver. Defaults to 1e-4.
maxiter: int, optional
    The maximum number of iterations. Defaults to 100.

Returns
-------
gains : $(array_type)
    Gain solutions of shape :code:`(time, ant, chan, dir, corr)`
    or shape :code:`(time, ant, chan, dir, corr, corr)`
jhj : $(array_type)
    The diagonal of the Hessian of shape
    :code:`(time, ant, chan, dir, corr)` or shape
    :code:`(time, ant, chan, dir, corr, corr)`
jhr : $(array_type)
    Residuals projected into gain space
    of shape :code:`(time, ant, chan, dir, corr)`
    or shape :code:`(time, ant, chan, dir, corr, corr)`.
k : int
    Number of iterations (will equal maxiter if
    not converged)
""")


try:
    phase_only_gauss_newton.__doc__ = PHASE_CALIBRATION_DOCS.substitute(
                                    array_type=":class:`numpy.ndarray`")
except AttributeError:
    pass

JHJ_AND_JHR_DOCS = DocstringTemplate("""
Computes the diagonal of the Hessian and
the residual projected in to gain space.
These are the terms required to perform
phase-only maximum likelihood calibration
assuming scalar or diagonal inputs.

Parameters
----------
time_bin_indices : $(array_type)
    The start indices of the time bins
    of shape :code:`(utime)`
time_bin_counts : $(array_type)
    The counts of unique time in each
    time bin of shape :code:`(utime)`
antenna1 : $(array_type)
    First antenna indices of shape :code:`(row,)`.
antenna2 : $(array_type)
    Second antenna indices of shape :code:`(row,)`
jones : $(array_type)
    Gain solutions of shape :code:`(time, ant, chan, dir, corr)`
    or :code:`(time, ant, chan, dir, corr, corr)`.
residual : $(array_type)
    Residual values of shape :code:`(row, chan, corr)`.
    or :code:`(row, chan, corr, corr)`.
model : $(array_type)
    Model data values of shape :code:`(row, chan, dir, corr)`
    or :code:`(row, chan, dir, corr, corr)`.
flag : $(array_type)
    Flag data of shape :code:`(row, chan, corr)`
    or :code:`(row, chan, corr, corr)`

Returns
-------
jhj : $(array_type)
    The diagonal of the Hessian of
    shape :code:`(time, ant, chan, dir, corr)`
    or :code:`(time, ant, chan, dir, corr, corr)`.
jhr : $(array_type)
    Residuals projected into signal space
    of shape :code:`(time, ant, chan, dir, corr)`
    or :code:`(time, ant, chan, dir, corr, corr)`.
""")


try:
    jhj_and_jhr.__doc__ = JHJ_AND_JHR_DOCS.substitute(
                                    array_type=":class:`numpy.ndarray`")
except AttributeError:
    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `codex-africanus` package."""

import numpy as np
import pytest

from africanus.model.coherency.conversion import (
    convert as np_convert)
from africanus.util.casa_types import STOKES_TYPE_MAP as smap

stokes_corr_cases = [
    ("complex", [['XX'], ['YY']],
     "real", ['I', 'Q']),
    ("complex", ['XX', 'YY'],
     "real", ['I', 'Q']),
    ("complex", ['XX', 'XY', 'YX', 'YY'],
     "real", ['I', 'Q', 'U', 'V']),
    ("complex", [['XX', 'XY'], ['YX', 'YY']],
     "real", [['I', 'Q'], ['U', 'V']]),
    ("real", ['I', 'Q', 'U', 'V'],
     "complex", ['XX', 'XY', 'YX', 'YY']),
    ("real", [['I', 'Q'], ['U', 'V']],
     "complex", [['XX', 'XY'], ['YX', 'YY']]),
    ("real", [['I', 'Q'], ['U', 'V']],
     "complex", [['XX', 'XY', 'YX', 'YY']]),
    ("real", [['I', 'Q'], ['U', 'V']],
     "complex", [['RR', 'RL', 'LR', 'LL']]),
    ("real", ['I', 'V'],
     "complex", ['RR', 'LL']),
    ("real", ['I', 'Q'],
     "complex", ['XX', 'YY']),
]

stokes_corr_int_cases = [
    ("complex", [smap['XX'], smap['YY']],
     "real", [smap['I'], smap['Q']])
]


vis_chunks = [
    ((10, 5, 3), (2, 3), (3,)),
    ((6, 8), (3, 3), (4, 4)),
    ((5, 5, 5),),
]


vis_shape = [tuple(sum(dim_chunks) for dim_chunks in case)
             for case in vis_chunks]


def visibility_factory(vis_shape, input_shape, in_type,
                       backend="numpy", **kwargs):
    shape = vis_shape + input_shape

    if backend == "numpy":
        vis = np.arange(1.0, np.product(shape) + 1.0)
        vis = vis.reshape(shape)
    elif backend == "dask":
        da = pytest.importorskip('dask.array')
        vis = da.arange(1.0, np.product(shape) + 1.0, chunks=np.product(shape))
        vis = vis.reshape(shape)
        vis = vis.rechunk(kwargs['vis_chunks'] + input_shape)
    else:
        raise ValueError("Invalid backend %s" % backend)

    if in_type == "real":
        pass
    elif in_type == "complex":
        vis = vis + 1j*vis
    else:
        raise ValueError("Invalid in_type %s" % in_type)

    return vis


@pytest.mark.parametrize("in_type, input_schema, out_type, output_schema",
                         stokes_corr_cases + stokes_corr_int_cases)
@pytest.mark.parametrize("vis_shape", vis_shape)
def test_conversion_schemas(in_type, input_schema,
                            out_type, output_schema,
                            vis_shape):
    input_shape = np.asarray(input_schema).shape
    output_shape = np.asarray(output_schema).shape
    vis = visibility_factory(vis_shape, input_shape, in_type)
    xformed_vis = np_convert(vis, input_schema, output_schema)
    assert xformed_vis.shape == vis_shape + output_shape


def test_conversion():
    I, Q, U, V = [1.0, 2.0, 3.0, 4.0]

    # Check conversion to linear (string)
    vis = np_convert(np.asarray([[I, Q, U, V]]),
                     ['I', 'Q', 'U', 'V'],
                     ['XX', 'XY', 'YX', 'YY'])

    XX, XY, YX, YY = vis[0]
    assert np.all(vis == [[I + Q, U + V*1j, U - V*1j, I - Q]])

    # Check conversion to linear (integer)
    vis = np_convert(np.asarray([[I, Q, U, V]]),
                     [smap[x] for x in ('I', 'Q', 'U', 'V')],
                     [smap[x] for x in ('XX', 'XY', 'YX', 'YY')])

    assert np.all(vis == [[I + Q, U + V*1j, U - V*1j, I - Q]])

    # Check conversion to circular (string)
    vis = np_convert(np.asarray([[I, Q, U, V]]),
                     ['I', 'Q', 'U', 'V'],
                     ['RR', 'RL', 'LR', 'LL'])

    RR, RL, LR, LL = vis[0]
    assert np.all(vis == [[I + V, Q + U*1j, Q - U*1j, I - V]])

    # Check conversion to circular (integer)
    vis = np_convert(np.asarray([[I, Q, U, V]]),
                     [smap[x] for x in ('I', 'Q', 'U', 'V')],
                     [smap[x] for x in ('RR', 'RL', 'LR', 'LL')])

    assert np.all(vis == [[I + V, Q + U*1j, Q - U*1j, I - V]])

    # linear to stokes (string)
    stokes = np_convert(np.asarray([[XX, XY, YX, YY]]),
                        ['XX', 'XY', 'YX', 'YY'],
                        ['I', 'Q', 'U', 'V'])

    assert np.all(stokes == [[I, Q, U, V]])

    # linear to stokes  (integer)
    stokes = np_convert(np.asarray([[XX, XY, YX, YY]]),
                        [smap[x] for x in ('XX', 'XY', 'YX', 'YY')],
                        [smap[x] for x in ('I', 'Q', 'U', 'V')])

    assert np.all(stokes == [[I, Q, U, V]])

    # circular to stokes (string)
    stokes = np_convert(np.asarray([[RR, RL, LR, LL]]),
                        ['RR', 'RL', 'LR', 'LL'],
                        ['I', 'Q', 'U', 'V'])

    assert np.all(stokes == [[I, Q, U, V]])

    # circular to stokes (intger)
    stokes = np_convert(np.asarray([[RR, RL, LR, LL]]),
                        [smap[x] for x in ('RR', 'RL', 'LR', 'LL')],
                        [smap[x] for x in ('I', 'Q', 'U', 'V')])

    assert np.all(stokes == [[I, Q, U, V]])


@pytest.mark.parametrize("in_type, input_schema, out_type, output_schema",
                         stokes_corr_cases + stokes_corr_int_cases)
@pytest.mark.parametrize("vis_chunks", vis_chunks)
def test_dask_conversion(in_type, input_schema,
                         out_type, output_schema,
                         vis_chunks):
    from africanus.model.coherency.dask import convert as da_convert

    vis_shape = tuple(sum(dim_chunks) for dim_chunks in vis_chunks)
    input_shape = np.asarray(input_schema).shape
    vis = visibility_factory(vis_shape, input_shape, in_type,
                             backend="dask", vis_chunks=vis_chunks)

    da_vis = da_convert(vis, input_schema, output_schema)
    np_vis = np_convert(vis.compute(), input_schema, output_schema)
    assert np.all(da_vis == np_vis)



# -*- coding: utf-8 -*-


from collections import namedtuple

import numpy as np
import numba

from africanus.averaging.support import unique_time, unique_baselines
from africanus.util.numba import is_numba_type_none, generated_jit, njit, jit


class RowMapperError(Exception):
    pass


def is_flagged_factory(have_flag_row):
    if have_flag_row:
        def impl(flag_row, r):
            return flag_row[r] != 0
    else:
        def impl(flag_row, r):
            return False

    return njit(nogil=True, cache=True)(impl)


def output_factory(have_flag_row):
    if have_flag_row:
        def impl(rows, flag_row):
            return np.zeros(rows, dtype=flag_row.dtype)
    else:
        def impl(rows, flag_row):
            return None

    return njit(nogil=True, cache=True)(impl)


def set_flag_row_factory(have_flag_row):
    if have_flag_row:
        def impl(flag_row, in_row, out_flag_row, out_row, flagged):
            if flag_row[in_row] == 0 and flagged:
                raise RowMapperError("Unflagged input row contributing "
                                     "to flagged output row. "
                                     "This should never happen!")

            out_flag_row[out_row] = (1 if flagged else 0)
    else:
        def impl(flag_row, in_row, out_flag_row, out_row, flagged):
            pass

    return njit(nogil=True, cache=True)(impl)


RowMapOutput = namedtuple("RowMapOutput",
                          ["map", "time", "interval", "flag_row"])


@generated_jit(nopython=True, nogil=True, cache=True)
def row_mapper(time, interval, antenna1, antenna2,
               flag_row=None, time_bin_secs=1):
    """
    Generates a mapping from a high resolution row index to
    a low resolution row index in support of time and channel
    averaging code. The `time` and `interval` columns
    are also respectively averaged and summed
    in the process of creating the mapping and a
    `flag_row` column is returned if one is provided.

    In order to average a chunk of row data, it is necessary to
    group each row (or sample) by baseline and then average
    the time samples present in each baseline in bins of
    `time_bin_secs`.

    Flagged data is handled as follows:

    1. It does not contribute to a bin at all if
       there are other unflagged samples in the bin.
    2. It is the only contribution to a bin if
       all samples in the bin are flagged.

    The algorithm is robust in the presence of missing time and baseline data.

    The algorithm works as follows:

    1. `time`, `interval`, `antenna1` and `antenna2`
    are used to construct a `row_lookup` array of shape `(ubl, utime)`
    mapping a baseline and time to a row of input data.

    2. For each baseline, `time_bin_secs` times are averaged together
    into two separate `time_lookup` arrays of shape `(ubl, utime)`.
    The first contains the average of unflagged samples, while the
    second contains the average of flagged samples.

    If the bin contains some unflagged samples, the unflagged average
    is used as the bin average, whereas if all samples are flagged
    the flagged average is used.

    Not all bins may be filled for a baseline if data is missing --
    these bins are assigned a sentinel value set to the
    maximum floating point value.

    A secondary `bin_lookup` array of shape `(ubl, utime)` is constructed
    mapping a time in the `row_lookup` array to a
    time bin in `time_lookup`.

    3. The `time_lookup` array is flattened and argsorted with a stable
    merge sort. As missing values are set to the maximum floating point
    value, this moves valid data to the front and missing data to the back.
    This has the effect of lexicographically sorts the data
    in an ascending `(time, bl)` order

    4. Input rows are then mapped via the `row_lookup`, `bin_lookup`
    and argsorted `time_lookup` arrays to an output row.

    .. code-block:: python

        ret = row_mapper(time, interval,
                         ant1, ant2, flag_row,
                         time_bin_secs=3)

        # Only add a bin's contribution if both input and output
        # are (a) flagged or (b) unflagged
        sel = flag_row == ret.flag_row[ret.map]
        sel_map = ret.map[sel]

        # Recompute time average using row map
        time = np.zeros_like(ret.time)
        ant1_avg = np.empty(time.shape, ant1.dtype)
        ant2_avg = np.empty(time.shape, ant2.dtype)
        counts = np.empty(time.shape, np.uint32)

        # Add time and 1 at map indices to time and counts
        np.add.at(time, sel_map, time[sel])
        np.add.at(counts, sel_map, 1)
        # Normalise
        time /= count

        np.testing.assert_array_equal(time, ret.time)

        # We assign baselines because each input baseline
        # is mapped to the same output baseline
        ant1_avg[sel_map] = ant1[sel]
        ant2_avg[sel_map] = ant2[sel]

    Parameters
    ----------
    time : :class:`numpy.ndarray`
        Time values of shape :code:`(row,)`.
    interval : :class:`numpy.ndarray`
        Exposure times of shape :code:`(row,)`.
    antenna1 : :class:`numpy.ndarray`
        Antenna 1 values of shape :code:`(row,)`.
    antenna2 : :class:`numpy.ndarray`
        Antenna 2 values of shape :code:`(row,)`.
    flag_row : :class:`numpy.ndarray`, optional
        Positive values indicate that a row is flagged, while
        zero implies unflagged. Has shape :code:`(row,)`.
    time_bin_secs : int, optional
        Number of timesteps to average into each bin

    Returns
    -------
    map : :class:`numpy.ndarray`
        Mapping from `np.arange(row)` to output row indices
        of shape :code:`(row,)`
    time : :class:`numpy.ndarray`
        Averaged time values of shape :code:`(out_row,)`
    interval : :class:`numpy.ndarray`
        Summed interval values of shape :code:`(out_row,)`
    flag_row : :class:`numpy.ndarray` or None
        Output flag rows of shape :code:`(out_row,)`.
        None if no input flag_row was supplied.

    Raises
    ------
    RowMapperError
        Raised if an illegal condition occurs

    """
    have_flag_row = not is_numba_type_none(flag_row)
    is_flagged_fn = is_flagged_factory(have_flag_row)

    output_flag_row = output_factory(have_flag_row)
    set_flag_row = set_flag_row_factory(have_flag_row)

    def impl(time, interval, antenna1, antenna2,
             flag_row=None, time_bin_secs=1):
        ubl, _, bl_inv, _ = unique_baselines(antenna1, antenna2)
        utime, _, time_inv, _ = unique_time(time)

        nbl = ubl.shape[0]
        ntime = utime.shape[0]

        sentinel = np.finfo(time.dtype).max
        out_rows = numba.uint32(0)

        scratch = np.full(3*nbl*ntime, -1, dtype=np.int32)
        row_lookup = scratch[:nbl*ntime].reshape(nbl, ntime)
        bin_lookup = scratch[nbl*ntime:2*nbl*ntime].reshape(nbl, ntime)
        inv_argsort = scratch[2*nbl*ntime:]
        time_lookup = np.zeros((nbl, ntime), dtype=time.dtype)
        interval_lookup = np.zeros((nbl, ntime), dtype=interval.dtype)

        # Is the entire bin flagged?
        bin_flagged = np.zeros((nbl, ntime), dtype=np.bool_)

        # Create a mapping from the full bl x time resolution back
        # to the original input rows
        for r in range(time.shape[0]):
            bl = bl_inv[r]
            t = time_inv[r]

            if row_lookup[bl, t] == -1:
                row_lookup[bl, t] = r
            else:
                raise ValueError("Duplicate (TIME, ANTENNA1, ANTENNA2) "
                                 "combinations were discovered in the input "
                                 "data. This is usually caused by not "
                                 "partitioning your data sufficiently "
                                 "by indexing columns, DATA_DESC_ID "
                                 "and SCAN_NUMBER in particular.")

        # Average times over each baseline and construct the
        # bin_lookup and time_lookup arrays
        for bl in range(ubl.shape[0]):
            tbin = numba.int32(0)
            bin_count = numba.int32(0)
            bin_flag_count = numba.int32(0)
            bin_low = time.dtype.type(0)

            for t in range(utime.shape[0]):
                # Lookup input row
                r = row_lookup[bl, t]

                # Ignore if not present
                if r == -1:
                    continue

                # At this point, we decide whether to contribute to
                # the current bin, or create a new one. We don't add
                # the current sample to the current bin if
                # high - low >= time_bin_secs
                half_int = interval[r] * 0.5

                # We're starting a new bin anyway,
                # just set the lower bin value
                if bin_count == 0:
                    bin_low = time[r] - half_int
                # If we exceed the seconds in the bin,
                # normalise the time and start a new bin
                elif time[r] + half_int - bin_low > time_bin_secs:
                    # Normalise and flag the bin
                    # if total counts match flagged counts
                    if bin_count > 0:
                        time_lookup[bl, tbin] /= bin_count
                        bin_flagged[bl, tbin] = bin_count == bin_flag_count
                    # There was nothing in the bin
                    else:
                        time_lookup[bl, tbin] = sentinel
                        bin_flagged[bl, tbin] = False

                    tbin += 1
                    bin_count = 0
                    bin_flag_count = 0

                # Record the output bin associated with the row
                bin_lookup[bl, t] = tbin

                # Time + Interval take unflagged + unflagged
                # samples into account (nominal value)
                time_lookup[bl, tbin] += time[r]
                interval_lookup[bl, tbin] += interval[r]
                bin_count += 1

                # Record flags
                if is_flagged_fn(flag_row, r):
                    bin_flag_count += 1

            # Normalise the last bin if it has entries in it
            if bin_count > 0:
                time_lookup[bl, tbin] /= bin_count
                bin_flagged[bl, tbin] = bin_count == bin_flag_count
                tbin += 1

            # Add this baseline's number of bins to the output rows
            out_rows += tbin

            # Set any remaining bins to sentinel value and unflagged
            for b in range(tbin, ntime):
                time_lookup[bl, b] = sentinel
                bin_flagged[bl, b] = False

        # Flatten the time lookup and argsort it
        flat_time = time_lookup.ravel()
        flat_int = interval_lookup.ravel()
        argsort = np.argsort(flat_time, kind='mergesort')

        # Generate lookup from flattened (bl, time) to output row
        for i, a in enumerate(argsort):
            inv_argsort[a] = i

        # Construct the final row map
        row_map = np.empty((time.shape[0]), dtype=np.uint32)

        # Construct output flag row, if necessary
        out_flag_row = output_flag_row(out_rows, flag_row)

        # foreach input row
        for in_row in range(time.shape[0]):
            # Lookup baseline and time
            bl = bl_inv[in_row]
            t = time_inv[in_row]

            # lookup time bin and output row
            tbin = bin_lookup[bl, t]
            # lookup output row in inv_argsort
            out_row = inv_argsort[bl*ntime + tbin]

            if out_row >= out_rows:
                raise RowMapperError("out_row >= out_rows")

            # Handle output row flagging
            set_flag_row(flag_row, in_row,
                         out_flag_row, out_row,
                         bin_flagged[bl, tbin])

            row_map[in_row] = out_row

        time_ret = flat_time[argsort[:out_rows]]
        int_ret = flat_int[argsort[:out_rows]]

        return RowMapOutput(row_map, time_ret, int_ret, out_flag_row)

    return impl


@jit(nopython=True, nogil=True, cache=True)
def channel_mapper(nchan, chan_bin_size=1):
    chan_map = np.empty(nchan, dtype=np.uint32)

    chan_bin = 0
    bin_count = 0

    for c in range(nchan):
        chan_map[c] = chan_bin
        bin_count += 1

        if bin_count == chan_bin_size:
            chan_bin += 1
            bin_count = 0

    if bin_count > 0:
        chan_bin += 1

    return chan_map, chan_bin



# -*- coding: utf-8 -*-


from africanus.model.spi.component_spi import SPI_DOCSTRING
from africanus.model.spi.component_spi import (
                                fit_spi_components as np_fit_spi_components)

from africanus.util.requirements import requires_optional

try:
    from dask.array.core import blockwise
except ImportError as e:
    opt_import_error = e
else:
    opt_import_error = None


def _fit_spi_components_wrapper(data, weights, freqs, freq0,
                                alphai, I0i, tol, maxiter):
    return np_fit_spi_components(data[0],
                                 weights[0],
                                 freqs[0],
                                 freq0,
                                 alphai[0] if alphai is not None else alphai,
                                 I0i[0] if I0i is not None else I0i,
                                 tol=tol,
                                 maxiter=maxiter)


@requires_optional('dask.array', opt_import_error)
def fit_spi_components(data, weights, freqs, freq0,
                       alphai=None, I0i=None,
                       tol=1e-5, maxiter=100):
    """ Dask wrapper fit_spi_components function """
    return blockwise(_fit_spi_components_wrapper, ("vars", "comps"),
                     data, ("comps", "chan"),
                     weights, ("chan",),
                     freqs, ("chan",),
                     freq0, None,
                     alphai, ("comps",) if alphai is not None else None,
                     I0i, ("comps",) if I0i is not None else None,
                     tol, None,
                     maxiter, None,
                     new_axes={"vars": 4},
                     dtype=data.dtype)


fit_spi_components.__doc__ = SPI_DOCSTRING.substitute(
                        array_type=":class:`dask.array.Array`")



# -*- coding: utf-8 -*-


from africanus.util.requirements import requires_optional


def register_assign_cycles(N, case=0):
    """
    Determine cycles that stem from performing
    an in-place assignment of the elements of an array.
    In the following. we cannot naively assign the source index
    to the dest index,
    If we assigned the value at index 0 to index 1, the assignment
    from index 1 to index 3 would be invalid, for example:

    src:  [3, 0, 2, 1]
    dest: [0, 1, 2, 3]

    assignment cycles can be broken by using a
    temporary variable to store the contents of a source index
    until dependent assignments have been performed.

    In this way, the minimum number of registers can be used
    to perform the in-place assignment.

    Returns
    -------
    list of lists of tuples
        For example, `[[(0, 2), (2, 0)], [1, 3], [3, 1]]`

    """
    dest = range(N)
    src = [(N - case + n) % N for n in dest]

    deps = {d: s for d, s in zip(dest, src) if d != s}

    for di, d in enumerate(dest):
        si = src.index(d)
        if si > di:
            deps[si] = di

    cycles = []

    while len(deps) > 0:
        k, v = deps.popitem()
        cycle = [(k, v)]

        while True:
            try:
                k = v
                v = deps.pop(k)
            except KeyError:
                # Check that the last key we're trying
                # to get is the first one in the cycle
                assert k == cycle[0][0]
                break

            cycle.append((k, v))

        cycles.append(cycle)

    return cycles


class TemplatingException(Exception):
    def __init__(self, msg):
        super(TemplatingException, self).__init__(msg)


def throw_helper(msg):
    raise TemplatingException(msg)


class FakeEnvironment(object):
    """
    Fake jinja2 environment, for which attribute/dict
    type access will fail
    """
    @requires_optional("jinja2")
    def __getitem__(self, key):
        raise NotImplementedError()

    @requires_optional("jinja2")
    def __setitem__(self, key, value):
        raise NotImplementedError()

    @requires_optional("jinja2")
    def __delitem__(self, key):
        raise NotImplementedError()

    @requires_optional("jinja2")
    def __getattr__(self, name):
        raise NotImplementedError()

    @requires_optional("jinja2")
    def __setattr__(self, name, value):
        raise NotImplementedError()

    @requires_optional("jinja2")
    def __delattr__(self, name):
        raise NotImplementedError()


def _jinja2_env_factory():
    try:
        from jinja2 import Environment, PackageLoader, select_autoescape
    except ImportError:
        return FakeEnvironment()

    loader = PackageLoader('africanus', '.')
    autoescape = select_autoescape(['j2', 'cu.j2'])
    env = Environment(loader=loader,
                      autoescape=autoescape,
                      extensions=['jinja2.ext.do'])

    # TODO(sjperkins)
    # Find a better way to set globals
    # perhaps search the package tree for e.g.
    # `jinja2_setup`.py files, whose contents
    # are inspected and assigned into the globals dict
    env.globals['register_assign_cycles'] = register_assign_cycles
    env.globals['throw'] = throw_helper

    return env


jinja_env = _jinja2_env_factory()



# -*- coding: utf-8 -*-


import logging
from os.path import join as pjoin

import numpy as np

from africanus.rime.feeds import FEED_ROTATION_DOCS
from africanus.util.code import format_code, memoize_on_key
from africanus.util.cuda import cuda_function, grids
from africanus.util.jinja2 import jinja_env
from africanus.util.requirements import requires_optional

try:
    import cupy as cp
    from cupy.core._scalar import get_typename as _get_typename
    from cupy.cuda.compiler import CompileException
except ImportError as e:
    opt_import_error = e
else:
    opt_import_error = None


log = logging.getLogger(__name__)


_TEMPLATE_PATH = pjoin("rime", "cuda", "feeds.cu.j2")


def _key_fn(parallactic_angles, feed_type):
    return (parallactic_angles.dtype, feed_type)


@memoize_on_key(_key_fn)
def _generate_kernel(parallactic_angles, feed_type):
    dtype = parallactic_angles.dtype

    # Block sizes
    if dtype == np.float32:
        block = (1024, 1, 1)
    elif dtype == np.float64:
        block = (512, 1, 1)
    else:
        raise TypeError("Unhandled type %s" % dtype)

    # Create template
    render = jinja_env.get_template(_TEMPLATE_PATH).render
    name = "feed_rotation"

    code = render(kernel_name=name,
                  feed_type=feed_type,
                  sincos_fn=cuda_function('sincos', dtype),
                  pa_type=_get_typename(dtype),
                  out_type=_get_typename(dtype))

    code = code.encode('utf-8')

    # Complex output type
    out_dtype = np.result_type(dtype, np.complex64)
    return cp.RawKernel(code, name), block, out_dtype


@requires_optional("cupy", opt_import_error)
def feed_rotation(parallactic_angles, feed_type='linear'):
    """ Cupy implementation of the feed_rotation kernel. """
    kernel, block, out_dtype = _generate_kernel(parallactic_angles, feed_type)
    in_shape = parallactic_angles.shape
    parallactic_angles = parallactic_angles.ravel()
    grid = grids((parallactic_angles.shape[0], 1, 1), block)
    out = cp.empty(shape=(parallactic_angles.shape[0], 4), dtype=out_dtype)

    try:
        kernel(grid, block, (parallactic_angles, out))
    except CompileException:
        log.exception(format_code(kernel.code))
        raise

    return out.reshape(in_shape + (2, 2))


try:
    feed_rotation.__doc__ = FEED_ROTATION_DOCS.substitute(
                                array_type=":class:`cupy.ndarray`")
except AttributeError:
    pass



# -*- coding: utf-8 -*-


import numpy as np
from africanus.util.docs import DocstringTemplate
from africanus.util.numba import generated_jit, njit

DIAG_DIAG = 0
DIAG = 1
FULL = 2


def check_type(jones, vis):
    """
    Determines which calibration scenario to apply i.e.
    DIAG_DIAG, DIAG or COMPLEX2x2.

    Parameters
    ----------
    jones : $(array_type)
        Jones term of shape :code:`(time, ant, chan, dir, corr)`
        or :code:`(time, ant, chan, dir, corr, corr)`
    vis : $(array_type)
        Visibility data of shape :code:`(row, chan, corr)`
        or :code:`(row, chan, corr, corr)`

    """
    vis_axes_count = vis.ndim
    jones_axes_count = jones.ndim
    if vis_axes_count == 3:
        mode = DIAG_DIAG
        if jones_axes_count != 5:
            raise RuntimeError("Jones axes not compatible with \
                                visibility axes. Expected length \
                                5 but got length %d" % jones_axes_count)

    elif vis_axes_count == 4:
        if jones_axes_count == 5:
            mode = DIAG

        elif jones_axes_count == 6:
            mode = FULL
        else:
            raise RuntimeError("Jones term has incorrect shape")
    else:
        raise RuntimeError("Visibility data has incorrect shape")

    return mode


def jones_inverse_mul_factory(mode):
    if mode == DIAG_DIAG:
        def jones_inverse_mul(a1j, blj, a2j, out):
            out[...] = blj/(a1j*np.conj(a2j))
    elif mode == DIAG:
        def jones_inverse_mul(a1j, blj, a2j, out):
            out[0, 0] = blj[0, 0]/(a1j[0]*np.conj(a2j[0]))
            out[0, 1] = blj[0, 1]/(a1j[0]*np.conj(a2j[1]))
            out[1, 0] = blj[1, 0]/(a1j[1]*np.conj(a2j[0]))
            out[1, 1] = blj[1, 1]/(a1j[1]*np.conj(a2j[1]))
    elif mode == FULL:
        def jones_inverse_mul(a1j, blj, a2j, out):
            # get determinant
            deta1j = a1j[0, 0]*a1j[1, 1]-a1j[0, 1]*a1j[1, 0]
            # compute inverse
            a00 = a1j[1, 1]/deta1j
            a01 = -a1j[1, 0]/deta1j
            a10 = -a1j[0, 1]/deta1j
            a11 = a1j[0, 0]/deta1j
            # get determinant
            deta2j = a2j[0, 0]*a2j[1, 1]-a2j[0, 1]*a2j[1, 0]
            # get conjugate transpose inverse
            b00 = np.conj(a2j[1, 1]/deta2j)
            b01 = np.conj(-a2j[1, 0]/deta2j)
            b10 = np.conj(-a2j[0, 1]/deta2j)
            b11 = np.conj(a2j[0, 0]/deta2j)
            # precompute resuable terms
            t1 = a00*blj[0, 0]
            t2 = a01*blj[1, 0]
            t3 = a00*blj[0, 1]
            t4 = a01*blj[1, 1]
            # overwrite with result
            out[0, 0] = t1*b00 +\
                t2*b00 +\
                t3*b10 +\
                t4*b10
            out[0, 1] = t1*b01 +\
                t2*b01 +\
                t3*b11 +\
                t4*b11
            t1 = a10*blj[0, 0]
            t2 = a11*blj[1, 0]
            t3 = a10*blj[0, 1]
            t4 = a11*blj[1, 1]
            out[1, 0] = t1*b00 +\
                t2*b00 +\
                t3*b10 +\
                t4*b10
            out[1, 1] = t1*b01 +\
                t2*b01 +\
                t3*b11 +\
                t4*b11
    return njit(nogil=True)(jones_inverse_mul)


def subtract_model_factory(mode):
    if mode == DIAG_DIAG:
        def subtract_model(a1j, blj, a2j, model, out):
            n_dir = np.shape(model)[0]
            out[...] = blj
            for s in range(n_dir):
                out -= a1j[s]*model[s]*np.conj(a2j[s])
    elif mode == DIAG:
        def subtract_model(a1j, blj, a2j, model, out):
            n_dir = np.shape(model)[0]
            out[...] = blj
            for s in range(n_dir):
                out[0, 0] -= a1j[s, 0]*model[s, 0, 0] * np.conj(a2j[s, 0])
                out[0, 1] -= a1j[s, 0]*model[s, 0, 1] * np.conj(a2j[s, 1])
                out[1, 0] -= a1j[s, 1]*model[s, 1, 0] * np.conj(a2j[s, 0])
                out[1, 1] -= a1j[s, 1]*model[s, 1, 1] * np.conj(a2j[s, 1])
    elif mode == FULL:
        def subtract_model(a1j, blj, a2j, model, out):
            n_dir = np.shape(model)[0]
            for s in range(n_dir):
                # precompute resuable terms
                t1 = a1j[s, 0, 0]*model[s, 0, 0]
                t2 = a1j[s, 0, 1]*model[s, 1, 0]
                t3 = a1j[s, 0, 0]*model[s, 0, 1]
                t4 = a1j[s, 0, 1]*model[s, 1, 1]
                tmp = np.conj(a2j[s].T)
                # overwrite with result
                out[0, 0] = blj[0, 0] -\
                    t1*tmp[0, 0] +\
                    t2*tmp[0, 0] +\
                    t3*tmp[1, 0] +\
                    t4*tmp[1, 0]
                out[0, 1] = blj[0, 1] -\
                    t1*tmp[0, 1] +\
                    t2*tmp[0, 1] +\
                    t3*tmp[1, 1] +\
                    t4*tmp[1, 1]
                t1 = a1j[s, 1, 0]*model[s, 0, 0]
                t2 = a1j[s, 1, 1]*model[s, 1, 0]
                t3 = a1j[s, 1, 0]*model[s, 0, 1]
                t4 = a1j[s, 1, 1]*model[s, 1, 1]
                out[1, 0] = blj[1, 0] -\
                    t1*tmp[0, 0] +\
                    t2*tmp[0, 0] +\
                    t3*tmp[1, 0] +\
                    t4*tmp[1, 0]
                out[1, 1] = blj[1, 1] -\
                    t1*tmp[0, 1] +\
                    t2*tmp[0, 1] +\
                    t3*tmp[1, 1] +\
                    t4*tmp[1, 1]
    return njit(nogil=True)(subtract_model)


@generated_jit(nopython=True, nogil=True, cache=True)
def correct_vis(time_bin_indices, time_bin_counts,
                antenna1, antenna2, jones, vis, flag):

    mode = check_type(jones, vis)

    jones_inverse_mul = jones_inverse_mul_factory(mode)

    def _correct_vis_fn(time_bin_indices, time_bin_counts,
                        antenna1, antenna2, jones, vis, flag):
        jones_shape = np.shape(jones)
        n_tim = jones_shape[0]
        n_dir = jones_shape[3]
        if n_dir > 1:
            raise ValueError("Jones has n_dir > 1.\
                                Cannot correct for direction dependent gains")
        n_chan = jones_shape[2]
        corrected_vis = np.zeros_like(vis, dtype=vis.dtype)
        for t in range(n_tim):
            for row in range(time_bin_indices[t],
                             time_bin_indices[t] + time_bin_counts[t]):
                p = int(antenna1[row])
                q = int(antenna2[row])
                gp = jones[t, p]
                gq = jones[t, q]
                for nu in range(n_chan):
                    if not np.any(flag[row, nu]):
                        jones_inverse_mul(
                            gp[nu, 0], vis[row, nu], gq[nu, 0],
                            corrected_vis[row, nu])
        return corrected_vis

    return _correct_vis_fn


@generated_jit(nopython=True, nogil=True, cache=True)
def residual_vis(time_bin_indices, time_bin_counts, antenna1,
                 antenna2, jones, vis, flag, model):

    mode = check_type(jones, vis)
    subtract_model = subtract_model_factory(mode)

    def _residual_vis_fn(time_bin_indices, time_bin_counts, antenna1,
                         antenna2, jones, vis, flag, model):
        n_tim = np.shape(time_bin_indices)[0]
        vis_shape = np.shape(vis)
        n_chan = vis_shape[1]
        residual = np.zeros(vis_shape, dtype=vis.dtype)
        for t in range(n_tim):
            for row in range(time_bin_indices[t],
                             time_bin_indices[t] + time_bin_counts[t]):
                p = int(antenna1[row])
                q = int(antenna2[row])
                gp = jones[t, p]
                gq = jones[t, q]
                for nu in range(n_chan):
                    if not np.any(flag[row, nu]):
                        subtract_model(
                            gp[nu], vis[row, nu], gq[nu],
                            model[row, nu], residual[row, nu])
        return residual

    return _residual_vis_fn


RESIDUAL_VIS_DOCS = DocstringTemplate("""
Computes residual visibilities in place
given model visibilities and gains solutions.

Parameters
----------
time_bin_indices : $(array_type)
    The start indices of the time bins
    of shape :code:`(utime)`
time_bin_counts : $(array_type)
    The counts of unique time in each
    time bin of shape :code:`(utime)`
antenna1 : $(array_type)
    First antenna indices of shape :code:`(row,)`.
antenna2 : $(array_type)
    Second antenna indices of shape :code:`(row,)`
jones : $(array_type)
    Gain solutions of shape :code:`(time, ant, chan, dir, corr)`
    or :code:`(time, ant, chan, dir, corr, corr)`.
vis : $(array_type)
    Data values of shape :code:`(row, chan, corr)`.
    or :code:`(row, chan, corr, corr)`.
flag : $(array_type)
    Flag data of shape :code:`(row, chan, corr)`
    or :code:`(row, chan, corr, corr)`
model : $(array_type)
    Model data values of shape :code:`(row, chan, dir, corr)`
    or :code:`(row, chan, dir, corr, corr)`.

Returns
-------
residual : $(array_type)
    Residual visibilities of shape
    :code:`(time, ant, chan, dir, corr)`
    or :code:`(time, ant, chan, dir, corr, corr)`.
""")


try:
    residual_vis.__doc__ = RESIDUAL_VIS_DOCS.substitute(
                                    array_type=":class:`numpy.ndarray`")
except AttributeError:
    pass

CORRECT_VIS_DOCS = DocstringTemplate("""
Apply DIE gains to visibilities to
generate corrected visibilities. For a
measurement model of the form

.. math::


    V_{pq} = G_{p} X_{pq} G_{q}^H + n_{pq}

the corrected visibilities are defined as

.. math::

    C_{pq} = G_{p}^{-1} V_{pq} G_{q}^{-H}

The corrected visibilities therefore have
a non-trivial noise contribution. Note
it is only possible to form corrected
data from direction independent gains
solutions so the :code:`dir` axis on
the jones terms should always be one.

Parameters
----------
time_bin_indices : $(array_type)
    The start indices of the time bins
    of shape :code:`(utime)`.
time_bin_counts : $(array_type)
    The counts of unique time in each
    time bin of shape :code:`(utime)`.
antenna1 : $(array_type)
    Antenna 1 index used to look up the antenna Jones
    for a particular baseline with shape :code:`(row,)`.
antenna2 : $(array_type)
    Antenna 2 index used to look up the antenna Jones
    for a particular baseline with shape :code:`(row,)`.
jones : $(array_type)
    Gain solutions of shape :code:`(time, ant, chan, dir, corr)`
    or :code:`(time, ant, chan, dir, corr, corr)`.
vis : $(array_type)
    Data values of shape :code:`(row, chan, corr)`
    or :code:`(row, chan, corr, corr)`.
flag : $(array_type)
    Flag data of shape :code:`(row, chan, corr)`
    or :code:`(row, chan, corr, corr)`.

Returns
-------
corrected_vis : $(array_type)
    True visibilities of shape :code:`(row,chan,corr_1,corr_2)`
""")

try:
    correct_vis.__doc__ = CORRECT_VIS_DOCS.substitute(
                        array_type=":class:`numpy.ndarray`")
except AttributeError:
    pass



import numpy as np
from numpy.testing import assert_array_equal, assert_array_almost_equal
import pytest


from africanus.rime.fast_beam_cubes import beam_cube_dde, freq_grid_interp


def rf(*a, **kw):
    return np.random.random(*a, **kw)


def rc(*a, **kw):
    return rf(*a, **kw) + 1j*rf(*a, **kw)


@pytest.fixture
def beam_freq_map():
    return np.array([.5, .56,  .7, .91, 1.0])


@pytest.fixture
def beam_freq_map_montblanc():
    """ Montblanc doesn't handle values outside the cube in the same way """
    return np.array([.4, .56,  .7, .91, 1.1])


@pytest.fixture
def freqs():
    """
    Related to the beam_freq_map fixture.
    Explanation of frequency test values:

    1. One value (0.4) below the beam freq range
    2. One value (0.5) on the first beam freq
    3. One value (1.0) on the last beam freq
    4. One value (1.1) above the beam freq range
    """

    return np.array([.4, .5, .6, .7, .8, .9, 1.0, 1.1])


def test_fast_beam_small():
    """ Small beam test, interpolation of one soure at [0.1, 0.1] """
    np.random.seed(42)

    # One frequency, to the lower side of the beam frequency map
    freq = np.asarray([.3])
    beam_freq_map = np.asarray([0.0, 1.0])

    beam_lw = 2
    beam_mh = 2
    beam_nud = beam_freq_map.shape[0]

    time = 1
    ants = 1
    chans = freq.shape[0]

    beam_extents = np.asarray([[-1.0, 1.0], [-1.0, 1.0]])
    lm = np.asarray([[0.1, 0.1]])

    # Set all of the following to identity
    parangles = np.zeros((time, ants))
    point_errors = np.zeros((time, ants, chans, 2))
    antenna_scaling = np.ones((ants, chans, 2))

    beam = rc((beam_lw, beam_mh, beam_nud, 1))

    ddes = beam_cube_dde(beam, beam_extents, beam_freq_map,
                         lm, parangles, point_errors, antenna_scaling,
                         freq)

    # Pen and paper the expected value
    lower_l = beam_extents[0, 0]
    lower_m = beam_extents[1, 0]

    ld = (lm[0, 0] - lower_l) / (beam_extents[0, 1] - lower_l)
    md = (lm[0, 1] - lower_m) / (beam_extents[1, 1] - lower_m)
    chd = freq[0]  # Hard-coded

    corr_sum = np.zeros((1,), dtype=beam.dtype)
    abs_sum = np.zeros((1,), dtype=beam.real.dtype)

    # Weights of the sample at each grid point
    wt0 = (1 - ld)*(1 - md)*(1 - chd)
    wt1 = ld*(1 - md)*(1 - chd)
    wt2 = (1 - ld)*md*(1 - chd)
    wt3 = ld*md*(1 - chd)

    wt4 = (1 - ld)*(1 - md)*chd
    wt5 = ld*(1 - md)*chd
    wt6 = (1 - ld)*md*chd
    wt7 = ld*md*chd

    # Sum lower channel correlations
    corr_sum[:] += wt0 * beam[0, 0, 0, 0]
    corr_sum[:] += wt1 * beam[1, 0, 0, 0]
    corr_sum[:] += wt2 * beam[0, 1, 0, 0]
    corr_sum[:] += wt3 * beam[1, 1, 0, 0]

    abs_sum[:] += wt0 * np.abs(beam[0, 0, 0, 0])
    abs_sum[:] += wt1 * np.abs(beam[1, 0, 0, 0])
    abs_sum[:] += wt2 * np.abs(beam[0, 1, 0, 0])
    abs_sum[:] += wt3 * np.abs(beam[1, 1, 0, 0])

    # Sum upper channel correlations
    corr_sum[:] += wt4 * beam[0, 0, 1, 0]
    corr_sum[:] += wt5 * beam[1, 0, 1, 0]
    corr_sum[:] += wt6 * beam[0, 1, 1, 0]
    corr_sum[:] += wt7 * beam[1, 1, 1, 0]

    abs_sum[:] += wt4 * np.abs(beam[0, 0, 1, 0])
    abs_sum[:] += wt5 * np.abs(beam[1, 0, 1, 0])
    abs_sum[:] += wt6 * np.abs(beam[0, 1, 1, 0])
    abs_sum[:] += wt7 * np.abs(beam[1, 1, 1, 0])

    corr_sum *= abs_sum / np.abs(corr_sum)

    assert_array_almost_equal([[[[[0.470255+0.4786j]]]]], ddes)
    assert_array_almost_equal(ddes.squeeze(), corr_sum.squeeze())


def test_grid_interpolate(freqs, beam_freq_map):
    freq_data = freq_grid_interp(freqs, beam_freq_map)

    freq_scale = freq_data[:, 0]
    fgrid_diff = freq_data[:, 1]
    grid_pos = np.int32(freq_data[:, 2])

    # Frequencies (first -- 0.8 and last -- 1.1)
    # outside the beam result in scaling,
    assert_array_almost_equal(freq_scale,
                              [0.8, 1.,  1.,  1.,  1.,  1.,  1.,  1.1])
    # Frequencies outside the beam are snapped to 0 if below
    # and beam_nud - 2 if above.
    # Frequencies on the edges are similarly snapped
    assert_array_equal(grid_pos, [0, 0, 1, 2, 2, 2, 3, 3])
    # Frequency less grid position frequency.
    # frequency is snapped to the first and last beam freq value
    # if outside (first and last values)
    # Third frequency value is also exactly on a grid point
    exp_diff = [1.,
                1.,
                0.71428571,
                1.,
                0.52380952,
                0.04761905,
                0.,
                0.]

    assert_array_almost_equal(fgrid_diff, exp_diff)


def test_dask_fast_beams(freqs, beam_freq_map):
    da = pytest.importorskip("dask.array")
    from africanus.rime.dask import beam_cube_dde as dask_beam_cube_dde

    beam_lw = 10
    beam_mh = 10
    beam_nud = beam_freq_map.shape[0]

    src_c = (2, 3, 5)
    time_c = (3, 2)
    ants_c = (2, 1, 1)
    chan_c = da.core.normalize_chunks(3, shape=freqs.shape)[0]

    src = sum(src_c)
    time = sum(time_c)
    ants = sum(ants_c)
    chans = sum(chan_c)

    # Source Transform variables
    lm = np.random.random(size=(src, 2))
    parangles = np.random.random(size=(time, ants))
    point_errors = np.random.random(size=(time, ants, chans, 2))
    antenna_scaling = np.random.random(size=(ants, chans, 2))

    # Make random values more representative
    lm = (lm - 0.5)*0.0001        # Shift lm to around the centre
    parangles *= np.pi / 12        # parangles to 15 degrees max
    point_errors *= 0.001         # Pointing errors
    antenna_scaling *= 0.0001     # Antenna scaling

    # Beam variables
    beam = rc((beam_lw, beam_mh, beam_nud, 2, 2))
    beam_lm_extents = np.asarray([[-1.0, 1.0], [-1.0, 1.0]])

    # Compute numba ddes
    ddes = beam_cube_dde(beam, beam_lm_extents, beam_freq_map,
                         lm, parangles, point_errors, antenna_scaling,
                         freqs)

    # Create dask arrays
    da_beam = da.from_array(beam, chunks=beam.shape)
    da_beam_freq_map = da.from_array(beam_freq_map, chunks=beam_freq_map.shape)
    da_lm = da.from_array(lm, chunks=(src_c, 2))
    da_parangles = da.from_array(parangles, chunks=(time_c, ants_c))
    da_point_errors = da.from_array(point_errors,
                                    chunks=(time_c, ants_c, chan_c, 2))
    da_ant_scale = da.from_array(antenna_scaling, chunks=(ants_c, chan_c, 2))
    da_extents = da.from_array(beam_lm_extents, chunks=beam_lm_extents.shape)
    da_freqs = da.from_array(freqs, chunks=chan_c)

    # dask ddes
    da_ddes = dask_beam_cube_dde(da_beam, da_extents, da_beam_freq_map,
                                 da_lm, da_parangles, da_point_errors,
                                 da_ant_scale, da_freqs)

    # Should be strictly equal
    assert_array_equal(da_ddes.compute(), ddes)


@pytest.mark.parametrize("dtype", [np.float32, np.float64])
def test_fast_beams_vs_montblanc(freqs, beam_freq_map_montblanc, dtype):
    """ Test that the numba beam matches montblanc implementation """
    mb_tf_mod = pytest.importorskip("montblanc.impl.rime.tensorflow")
    tf = pytest.importorskip("tensorflow")

    freqs = freqs.astype(dtype)
    beam_freq_map = beam_freq_map_montblanc.astype(dtype)

    beam_lw = 10
    beam_mh = 10
    beam_nud = beam_freq_map.shape[0]

    src = 10
    time = 5
    ants = 4
    chans = freqs.shape[0]

    ctype = np.result_type(dtype, np.complex64)

    # Source Transform variables
    lm = np.random.random(size=(src, 2)).astype(dtype)
    parangles = np.random.random(size=(time, ants)).astype(dtype)
    point_errors = np.random.random(size=(time, ants, chans, 2)).astype(dtype)
    antenna_scaling = np.random.random(size=(ants, chans, 2)).astype(dtype)

    # Make random values more representative
    lm = (lm - 0.5)*0.0001        # Shift lm to around the centre
    parangles *= np.pi / 12        # parangles to 15 degrees max
    point_errors *= 0.001         # Pointing errors
    antenna_scaling *= 0.0001     # Antenna scaling

    # Beam variables
    beam = rc((beam_lw, beam_mh, beam_nud, 2, 2)).astype(ctype)
    beam_lm_extents = np.asarray([[-1.0, 1.0], [-1.0, 1.0]]).astype(dtype)

    ddes = beam_cube_dde(beam, beam_lm_extents, beam_freq_map,
                         lm, parangles, point_errors, antenna_scaling,
                         freqs)

    assert ddes.shape == (src, time, ants, chans, 2, 2)

    rime = mb_tf_mod.load_tf_lib()

    # Montblanc beam extent format is different
    mb_beam_extents = np.array([beam_lm_extents[0, 0],
                                beam_lm_extents[1, 0],
                                beam_freq_map[0],
                                beam_lm_extents[0, 1],
                                beam_lm_extents[1, 1],
                                beam_freq_map[-1]], dtype=dtype)

    # Montblanc wants flattened correlations
    mb_beam = beam.reshape(beam.shape[:3] + (-1,))

    np_args = [lm, freqs, point_errors, antenna_scaling,
               np.sin(parangles), np.cos(parangles),
               mb_beam_extents, beam_freq_map, mb_beam]
    # Argument string name list
    arg_names = ["lm", "frequency", "point_errors", "antenna_scaling",
                 "parallactic_angle_sin", "parallactic_angle_cos",
                 "beam_extents", "beam_freq_map", "e_beam"]

    # Constructor tensorflow variables
    tf_args = [tf.Variable(v, name=n) for v, n in zip(np_args, arg_names)]

    with tf.device("/cpu:0"):
        tf_dde = rime.e_beam(*tf_args)

    init_op = tf.global_variables_initializer()

    with tf.Session() as S:
        S.run(init_op)
        res = S.run(tf_dde)

    assert_array_almost_equal(res, ddes.reshape(ddes.shape[:4] + (-1,)))



# -*- coding: utf-8 -*-


import collections

import numpy as np

from .kaiser_bessel_filter import (kaiser_bessel_with_sinc,
                                   estimate_kaiser_bessel_beta)

ConvolutionFilter = collections.namedtuple("ConvolutionFilter",
                                           ['half_sup', 'oversample',
                                            'full_sup_wo_padding', 'full_sup',
                                            'no_taps', 'filter_taps'])
"""
:class:`collections.namedtuple` containing attributes
defining a 2D Convolution Filter. A namedtuple is used
because they're easier to use when using
``nopython`` mode in :mod:`numba`.

.. attribute:: full_sup

    Full support

.. attribute:: full_sup_wo_padding

    Full support without padding

.. attribute:: half_sup

    Half support

.. attribute:: oversample

    Oversampling factor

.. attribute:: no_taps

    Number of taps

.. attribute:: filter_taps

    2D filter taps with shape (v, u)
"""


class AsymmetricKernel(Exception):
    pass


def convolution_filter(half_support, oversampling_factor,
                       filter_type, **kwargs):
    r"""
    Create a 2D Convolution Filter suitable
    for use with gridding and degridding functions.

    Parameters
    ----------
    half_support : integer
        Half support (N) of the filter. The filter has a
        full support of N*2 + 3 taps.
        Two of the taps exist as padding.
    oversampling_factor : integer
        Number of spaces in-between grid-steps
        (improves gridding/degridding accuracy)
    filter_type : {'kaiser-bessel', 'sinc'}
        Filter type. See `Convolution Filters <convolution-filter-api_>`_
        for further information.
    beta : float, optional
        Beta shape parameter for
        `Kaiser Bessel <kaiser-bessel-filter_>`_ filters.
    normalise : {True, False}
        Normalise the filter by the it's volume.
        Defaults to ``True``.

    Returns
    -------
    :class:`ConvolutionFilter`
        namedtuple containing filter attributes
    """
    full_sup_wo_padding = (half_support * 2 + 1)
    full_sup = full_sup_wo_padding + 2  # + padding
    no_taps = full_sup + (full_sup - 1) * (oversampling_factor - 1)

    normalise = kwargs.pop("normalise", True)

    taps = np.arange(no_taps) / oversampling_factor - full_sup // 2

    if filter_type == 'sinc':
        filter_taps = np.sinc(taps)
    elif filter_type == 'kaiser-bessel':
        # https://www.dsprelated.com/freebooks/sasp/Kaiser_Window.html
        try:
            beta = kwargs.pop('beta')
        except KeyError:
            beta = estimate_kaiser_bessel_beta(full_sup)

        # Compute Kaiser Bessel and multiply in the sinc
        filter_taps = kaiser_bessel_with_sinc(taps, full_sup,
                                              oversampling_factor, beta,
                                              normalise=normalise)
    else:
        raise ValueError("Expected one of {'kaiser-bessel', 'sinc'}")

    # Expand filter taps to 2D
    filter_taps = np.outer(filter_taps, filter_taps)

    if not np.all(filter_taps == filter_taps.T):
        raise AsymmetricKernel("Kernel is asymmetric")

    return ConvolutionFilter(half_support, oversampling_factor,
                             full_sup_wo_padding, full_sup,
                             no_taps, filter_taps)



# -*- coding: utf-8 -*-


import numpy as np
from numpy.testing import assert_array_almost_equal
from africanus.dft import im_to_vis
from africanus.averaging.support import unique_time
from africanus.calibration import phase_only_gauss_newton
from africanus.rime.predict import predict_vis

n_time = 32
n_chan = 16
n_ant = 7
n_bl = n_ant*(n_ant-1)//2
n_row = n_bl*n_time
n_cor = 2


def give_lm(n_dir):
    ls = 0.1*np.random.randn(n_dir)
    ms = 0.1*np.random.randn(n_dir)
    lm = np.vstack((ls, ms)).T
    return lm


def give_flux(n_dir):
    return 0.1 + np.random.random(n_dir)


def make_dual_pol_data(sigma_n, n_dir, sigma_f):
    # make aux data
    antenna1 = np.zeros(n_row, dtype=np.int16)
    antenna2 = np.zeros(n_row, dtype=np.int16)
    time = np.zeros(n_row, dtype=np.float32)
    uvw = np.zeros((n_row, 3), dtype=np.float32)
    unique_time = np.linspace(0, 1, n_time)
    freq = np.linspace(1e9, 2e9, n_chan)
    for i in range(n_time):
        row = 0
        for p in range(n_ant):
            for q in range(p):
                time[i*n_bl + row] = unique_time[i]
                antenna1[i*n_bl + row] = p
                antenna2[i*n_bl + row] = q
                uvw[i*n_bl + row] = np.random.randn(3)
                row += 1
    assert time.size == n_row
    # simulate visibilities
    model_data = np.zeros((n_row, n_chan, n_dir, n_cor), dtype=np.complex64)
    # make up some sources
    lm = give_lm(n_dir)
    flux = give_flux(n_dir)
    # simulate model data (pure Stokes I)
    for dir in range(n_dir):
        this_lm = lm[dir].reshape(1, 2)
        this_flux = np.tile(flux[dir], (n_chan, n_cor))[None, :, :]
        model_tmp = im_to_vis(this_flux, uvw, this_lm, freq)
        model_data[:, :, dir, :] = model_tmp
    assert not np.isnan(model_data).any()
    # simulate gains (just radnomly scattered around 1 for now)
    jones = np.ones((n_time, n_ant, n_chan, n_dir, n_cor), dtype=np.complex64)
    if sigma_f:
        jones += sigma_f*(
            np.random.randn(n_time, n_ant, n_chan, n_dir, n_cor) +
            1.0j*np.random.randn(n_time, n_ant, n_chan, n_dir, n_cor))
        assert (np.abs(jones) > 1e-5).all()
        assert not np.isnan(jones).any()
    # get vis
    time_index = np.unique(time, return_inverse=True)[1]
    jones_tmp = np.transpose(jones, [3, 0, 1, 2, 4])
    model_tmp = np.transpose(model_data, [2, 0, 1, 3])
    vis = predict_vis(
        time_index, antenna1, antenna2,
        source_coh=model_tmp,
        dde1_jones=jones_tmp,
        dde2_jones=jones_tmp)
    assert not np.isnan(vis).any()
    # add noise
    if sigma_n:
        vis += sigma_n*(np.random.randn(n_row, n_chan, n_cor) +
                        1.0j*np.random.randn(n_row, n_chan, n_cor))
    weights = np.ones((n_row, n_chan, n_cor), dtype=np.float32)
    if sigma_n:
        weights /= sigma_n**2
    flag = np.zeros((n_row, n_chan, n_cor), dtype=np.bool)
    data_dict = {}
    data_dict["DATA"] = vis
    data_dict["MODEL_DATA"] = model_data
    data_dict["WEIGHT_SPECTRUM"] = weights
    data_dict["TIME"] = time
    data_dict["ANTENNA1"] = antenna1
    data_dict["ANTENNA2"] = antenna2
    data_dict["FLAG"] = flag
    data_dict['JONES'] = jones
    return data_dict


def test_phase_only_diag_diag():
    """
    Test phase only calibration by checking that
    we reconstruct the correct gains for a noise
    free simulation.
    """
    np.random.seed(42)
    # simulate noise free data with random DDE's
    n_dir = 1
    sigma_n = 0.0
    sigma_f = 0.05
    data_dict = make_dual_pol_data(sigma_n, n_dir, sigma_f)
    time = data_dict['TIME']
    _, time_bin_indices, _, time_bin_counts = unique_time(time)
    ant1 = data_dict['ANTENNA1']
    ant2 = data_dict['ANTENNA2']
    vis = data_dict['DATA']
    model = data_dict['MODEL_DATA']
    jones = data_dict['JONES']
    flag = data_dict['FLAG']
    weight = data_dict['WEIGHT_SPECTRUM']
    # calibrate the data
    jones0 = np.ones((n_time, n_ant, n_chan, n_dir, n_cor), dtype=np.complex64)
    precision = 5
    gains, jhj, jhr, k = phase_only_gauss_newton(
        time_bin_indices, time_bin_counts,
        ant1, ant2, jones0, vis,
        flag, model, weight,
        tol=10**(-precision), maxiter=100)
    # check that phase differences are correct
    for p in range(n_ant):
        for q in range(p):
            phase_diff_true = np.angle(jones[:, p] * np.conj(jones[:, q]))
            phase_diff = np.angle(gains[:, p] * np.conj(gains[:, q]))
            assert_array_almost_equal(
                phase_diff_true, phase_diff, decimal=precision-1)



# -*- coding: utf-8 -*-


# NOTE(sjperkins)
# Non standard library imports should be avoided,
# or should fail gracefully as functionality
# in these modules is called by setup.py
import os

# requirements
on_rtd = os.environ.get('READTHEDOCS') == 'True'

# Basic requirements containing no C extensions.
# This is necessary for building on RTD
requirements = ['appdirs >= 1.4.3',
                'decorator']

if not on_rtd:
    requirements += [
        # astropy breaks with numpy 1.15.3
        # https://github.com/astropy/astropy/issues/7943
        'numpy >= 1.14.0, != 1.15.3',
        'numba >= 0.43.0']

extras_require = {
    'cuda': ['cupy >= 5.0.0', 'jinja2 >= 2.10'],
    'dask': ['dask[array] >= 1.1.0'],
    'jax': ['jax == 0.1.27', 'jaxlib == 0.1.14'],
    'scipy': ['scipy >= 1.0.0'],
    'astropy': ['astropy >= 2.0.0, < 3.0; python_version <= "2.7"',
                'astropy >= 3.0; python_version >= "3.0"'],
    'python-casacore': ['python-casacore == 3.0.0'],
    'testing': ['pytest', 'flaky', 'pytest-flake8']
}

_non_cuda_extras = [er for n, er in extras_require.items() if n != "cuda"]
_all_extras = extras_require.values()

extras_require['complete'] = sorted(set(sum(_non_cuda_extras, [])))
extras_require['complete-cuda'] = sorted(set(sum(_all_extras, [])))

setup_requirements = []
test_requirements = (extras_require['testing'] +
                     extras_require['astropy'] +
                     extras_require['python-casacore'] +
                     extras_require['dask'] +
                     extras_require['scipy'])



# -*- coding: utf-8 -*-


from functools import reduce
from operator import mul

from africanus.util.numba import jit

import numpy as np

_ARCSEC2RAD = np.deg2rad(1.0/(60*60))


@jit(nopython=True, nogil=True, cache=True)
def numba_grid(vis, uvw, flags, weights, ref_wave,
               convolution_filter, cell_size, grid):
    """
    See :func:"~africanus.gridding.simple.gridding.grid" for
    documentation.
    """
    cf = convolution_filter

    # Shape checks
    assert vis.shape[0] == uvw.shape[0] == flags.shape[0] == weights.shape[0]
    assert vis.shape[1] == flags.shape[1] == weights.shape[1]
    assert vis.shape[2] == flags.shape[2] == weights.shape[2]

    nrow, nchan = vis.shape[0:2]
    assert nchan == ref_wave.shape[0]
    corrs = vis.shape[2:]

    ny, nx = grid.shape[0:2]
    flat_corrs = grid.shape[2]

    # Similarity Theorem
    # https://www.cv.nrao.edu/course/astr534/FTSimilarity.html
    # Scale UV coordinates
    # Note u => x and v => y
    u_scale = _ARCSEC2RAD * cell_size * nx
    v_scale = _ARCSEC2RAD * cell_size * ny

    # Flatten correlation dimension for easier loop handling
    fvis = vis.reshape((nrow, nchan, flat_corrs))
    fflags = flags.reshape((nrow, nchan, flat_corrs))
    fweights = weights.reshape((nrow, nchan, flat_corrs))

    filter_index = np.arange(-cf.half_sup, cf.half_sup+1)

    half_x = nx // 2
    half_y = ny // 2

    for r in range(uvw.shape[0]):                 # row (vis)
        for f in range(vis.shape[1]):             # channel (freq)
            # Exact UV coordinates
            exact_u = uvw[r, 0] * u_scale / ref_wave[f]
            exact_v = uvw[r, 1] * v_scale / ref_wave[f]

            # Discretised UV coordinates
            disc_u = int(np.round(exact_u))
            disc_v = int(np.round(exact_v))

            extent_u = disc_u + half_x
            extent_v = disc_v + half_y

            # Out of bounds check
            if (extent_v + cf.half_sup >= ny or
                extent_u + cf.half_sup >= nx or
                extent_v - cf.half_sup < 0 or
                    extent_u - cf.half_sup < 0):
                continue

            # One plus half support (our kernels have 1 pixel of extra padding)
            one_half_sup = 1 + cf.half_sup

            # Compute fractional u and v
            base_frac_u = disc_u - exact_u
            base_frac_v = disc_v - exact_v

            frac_u = int(np.round(base_frac_u*cf.oversample))
            frac_v = int(np.round(base_frac_v*cf.oversample))

            # Iterate over v/y
            for conv_v in filter_index:
                v_idx = (conv_v + one_half_sup)*cf.oversample + frac_v
                grid_v = disc_v + conv_v + half_y

                # Iterate over u/x
                for conv_u in filter_index:
                    u_idx = (conv_u + one_half_sup)*cf.oversample + frac_u
                    conv_weight = cf.filter_taps[v_idx, u_idx]
                    grid_u = disc_u + conv_u + half_x

                    for c in range(flat_corrs):      # correlation
                        # Ignore flagged correlations
                        if fflags[r, f, c] > 0:
                            continue

                        # Grid the visibility
                        grid[grid_v, grid_u, c] += (fvis[r, f, c] *
                                                    conv_weight *
                                                    fweights[r, f, c])

    return grid.reshape((ny, nx) + corrs)


def grid(vis, uvw, flags, weights, ref_wave,
         convolution_filter,
         cell_size,
         nx=1024, ny=1024,
         grid=None):
    """
    Convolutional gridder which grids visibilities ``vis``
    at the specified ``uvw`` coordinates and
    ``ref_wave`` reference wavelengths using
    the specified ``convolution_filter``.

    Variable numbers of correlations are supported.

    * :code:`(row, chan, corr_1, corr_2)` ``vis`` will result in a
      :code:`(ny, nx, corr_1, corr_2)` ``grid``.
    * :code:`(row, chan, corr_1)` ``vis`` will result in a
      :code:`(ny, nx, corr_1)` ``grid``.

    Parameters
    ----------
    vis : np.ndarray
        complex visibility array of shape :code:`(row, chan, corr_1, corr_2)`
    uvw : np.ndarray
        float64 array of UVW coordinates of shape :code:`(row, 3)`
        in wavelengths.
    weights : np.ndarray
        float32 or float64 array of weights. Set this to
        ``np.ones_like(vis, dtype=np.float32)`` as default.
    flags : np.ndarray
        flagged array of shape :code:`(row, chan, corr_1, corr_2)`.
        Any positive quantity will indicate that the corresponding
        visibility should be flagged.
        Set to ``np.zeros_like(vis, dtype=np.bool)`` as default.
    ref_wave : np.ndarray
        float64 array of wavelengths of shape :code:`(chan,)`
    convolution_filter :  :class:`~africanus.filters.ConvolutionFilter`
        Convolution filter
    cell_size : float
        Cell size in arcseconds.
    nx : integer, optional
        Size of the grid's X dimension
    ny : integer, optional
        Size of the grid's Y dimension
    grid : np.ndarray, optional
        complex64/complex128 array of shape :code:`(ny, nx, corr_1, corr_2)`
        If supplied, this array will be used as the gridding target,
        and ``nx`` and ``ny`` will be derived from this grid's
        dimensions.

    Returns
    -------
    np.ndarray
        :code:`(ny, nx, corr_1, corr_2)` complex ndarray of
        gridded visibilities. The number of correlations may vary,
        depending on the shape of vis.
    """

    # Flatten the correlation dimensions
    corrs = vis.shape[2:]
    flat_corrs = (reduce(mul, corrs),)

    # Create grid of flatten correlations or reshape
    if grid is None:
        grid = np.zeros((ny, nx) + flat_corrs, dtype=vis.dtype)
    else:
        ny, nx = grid.shape[0:2]
        grid = grid.reshape((ny, nx) + flat_corrs)

    return numba_grid(vis, uvw, flags, weights, ref_wave,
                      convolution_filter, cell_size, grid)


@jit(nopython=True, nogil=True, cache=True)
def numba_degrid(grid, uvw, weights, ref_wave,
                 convolution_filter, cell_size, vis):
    """
    See :func:"~africanus.gridding.simple.gridding.degrid" for
    documentation.
    """
    cf = convolution_filter
    ny, nx, flat_corrs = grid.shape

    # Similarity Theorem
    # https://www.cv.nrao.edu/course/astr534/FTSimilarity.html
    # Scale UV coordinates
    # Note u => x and v => y
    u_scale = _ARCSEC2RAD * cell_size * nx
    v_scale = _ARCSEC2RAD * cell_size * ny

    filter_index = np.arange(-cf.half_sup, cf.half_sup+1)

    half_x = nx // 2
    half_y = ny // 2

    for r in range(uvw.shape[0]):                 # row (vis)
        for f in range(vis.shape[1]):             # channel (freq)
            exact_u = uvw[r, 0] * u_scale / ref_wave[f]
            exact_v = uvw[r, 1] * v_scale / ref_wave[f]

            disc_u = int(np.round(exact_u))
            disc_v = int(np.round(exact_v))

            extent_v = disc_v + half_y
            extent_u = disc_u + half_x

            # Out of bounds check
            if (extent_v + cf.half_sup >= ny or
                extent_u + cf.half_sup >= nx or
                extent_v - cf.half_sup < 0 or
                    extent_u - cf.half_sup < 0):
                continue

            # One plus half support
            one_half_sup = 1 + cf.half_sup

            # Compute fractional u and v
            base_frac_u = disc_u - exact_u
            base_frac_v = disc_v - exact_v

            frac_u = int(np.round(base_frac_u*cf.oversample))
            frac_v = int(np.round(base_frac_v*cf.oversample))

            # Iterate over v/y
            for conv_v in filter_index:
                v_idx = (conv_v + one_half_sup)*cf.oversample + frac_v
                grid_v = disc_v + conv_v + half_y

                # Iterate over u/x
                for conv_u in filter_index:
                    u_idx = (conv_u + one_half_sup)*cf.oversample + frac_u
                    conv_weight = cf.filter_taps[v_idx, u_idx]
                    grid_u = disc_u + conv_u + half_x

                    # Correlation
                    for c in range(flat_corrs):
                        vis[r, f, c] += (grid[grid_v, grid_u, c] *
                                         conv_weight *
                                         weights[r, f, c])

    return vis


def degrid(grid, uvw, weights, ref_wave,
           convolution_filter, cell_size, dtype=np.complex64):
    """
    Convolutional degridder (continuum)

    Variable numbers of correlations are supported.

    * :code:`(ny, nx, corr_1, corr_2)` ``grid`` will result in a
      :code:`(row, chan, corr_1, corr_2)` ``vis``

    * :code:`(ny, nx, corr_1)` ``grid`` will result in a
      :code:`(row, chan, corr_1)` ``vis``

    Parameters
    ----------
    grid : np.ndarray
        float or complex grid of visibilities
        of shape :code:`(ny, nx, corr_1, corr_2)`
    uvw : np.ndarray
        float64 array of UVW coordinates of shape :code:`(row, 3)`
        in wavelengths.
    weights : np.ndarray
        float32 or float64 array of weights. Set this to
        ``np.ones_like(vis, dtype=np.float32)`` as default.
    ref_wave : np.ndarray
        float64 array of wavelengths of shape :code:`(chan,)`
    convolution_filter :  :class:`~africanus.filters.ConvolutionFilter`
        Convolution Filter
    cell_size : float
        Cell size in arcseconds.
    dtype : :class:`numpy.dtype`
        Data type of the visibilities

    Returns
    -------
    np.ndarray
        :code:`(row, chan, corr_1, corr_2)` complex ndarray of visibilities
    """
    nrow = uvw.shape[0]
    nchan = ref_wave.shape[0]
    corrs = flat_corrs = grid.shape[2:]

    # Flatten if necessary
    if len(corrs) > 1:
        flat_corrs = (reduce(mul, corrs),)
        grid = grid.reshape(grid.shape[:2] + flat_corrs)
        weights = weights.reshape(weights.shape[:2] + flat_corrs)

    vis = np.zeros((nrow, nchan) + flat_corrs, dtype=dtype)

    vis = numba_degrid(grid, uvw, weights, ref_wave,
                       convolution_filter, cell_size, vis)

    return vis.reshape(weights.shape[:2] + corrs)



# -*- coding: utf-8 -*-


import numpy as np

from .kaiser_bessel_filter import (kaiser_bessel_with_sinc,
                                   estimate_kaiser_bessel_beta)


def taper(filter_type, ny, nx, conv_filter, **kwargs):
    r"""
    Parameters
    ----------
    filter_type : {"kaiser-bessel"}
        Type of filter
    ny : int
        Number of pixels in the v dimension.
    nx : int
        Number of pixels in the u dimension.
    conv_filter : :class:`africanus.filters.ConvolutionFilter`
        Associated Convolution Filter.

    Returns
    -------
    :class:`numpy.ndarray`
        Taper of shape :code:`(ny, nx)`
    """
    cf = conv_filter

    if filter_type == "sinc":
        return np.ones((ny, nx))
    elif filter_type == "kaiser-bessel":
        try:
            beta = kwargs.pop('beta')
        except KeyError:
            beta = estimate_kaiser_bessel_beta(cf.full_sup)

        # What would Andre Offringa do?
        # He would compute the numeric solution
        taps = np.arange(cf.no_taps) / cf.oversample - cf.full_sup // 2
        kb = kaiser_bessel_with_sinc(taps, cf.full_sup, cf.oversample, beta)
        kbshift = np.fft.fftshift(kb)

        width = nx * cf.oversample
        height = ny * cf.oversample

        # Put the first and last halves of the shifted Kaiser Bessel
        # at each end of the output buffer, then FFT
        buf = np.zeros(width, dtype=kb.dtype)
        buf[:kbshift.size // 2] = kbshift[:kbshift.size // 2]
        buf[-kbshift.size // 2:] = kbshift[-kbshift.size // 2:]
        x = np.fft.ifft(buf).real

        buf = np.zeros(height, dtype=kb.dtype)
        buf[:kbshift.size // 2] = kbshift[:kbshift.size // 2]
        buf[-kbshift.size // 2:] = kbshift[-kbshift.size // 2:]
        y = np.fft.ifft(buf).real

        # First quarter of the taper
        quarter = y[:ny // 2, None] * x[None, :nx // 2]

        # Create the taper by copying
        # the quarter into the appropriate bits
        taper = np.empty((ny, nx), dtype=kb.dtype)
        taper[:ny // 2, :nx // 2] = quarter[::-1, ::-1]
        taper[ny // 2:, :nx // 2] = quarter[:, ::-1]
        taper[:ny // 2, nx // 2:] = quarter[::-1, :]
        taper[ny // 2:, nx // 2:] = quarter[:, :]

        # Normalise by oversampling factor
        taper *= cf.oversample**2

        return taper
    else:
        raise ValueError("Invalid filter_type '%s'" % filter_type)



# -*- coding: utf-8 -*-


from numba import types
import numpy as np

from africanus.util.numba import generated_jit, njit
from africanus.util.docs import DocstringTemplate


def numpy_spectral_model(stokes, spi, ref_freq, frequency, base):
    out_shape = (stokes.shape[0], frequency.shape[0]) + stokes.shape[1:]

    # Add in missing pol dimensions
    if stokes.ndim == 1:
        stokes = stokes[:, None]

    if spi.ndim == 2:
        spi = spi[:, :, None]

    npol = spi.shape[2]

    if isinstance(base, list):
        base = base + [base[-1]] * (npol - len(base))
    else:
        base = [base]*npol

    spi_exps = np.arange(1, spi.shape[1] + 1)

    spectral_model = np.empty((stokes.shape[0], frequency.shape[0], npol),
                              dtype=stokes.dtype)

    for p, b in enumerate(base):
        if b in ("std", 0):
            freq_ratio = (frequency[None, :] / ref_freq[:, None]) - 1.0
            term = freq_ratio[:, None, :]**spi_exps[None, :, None]
            term = spi[:, :, p, None] * term
            spectral_model[:, :, p] = stokes[:, p, None] + term.sum(axis=1)
        elif b in ("log", 1):
            freq_ratio = np.log(frequency[None, :] / ref_freq[:, None])
            term = freq_ratio[:, None, :]**spi_exps[None, :, None]
            term = spi[:, :, p, None] * term
            spectral_model[:, :, p] = (np.exp(np.log(stokes[:, p, None]) +
                                              term.sum(axis=1)))
        elif b in ("log10", 2):
            freq_ratio = np.log10(frequency[None, :] / ref_freq[:, None])
            term = freq_ratio[:, None, :]**spi_exps[None, :, None]
            term = spi[:, :, p, None] * term
            spectral_model[:, :, p] = (10**(np.log10(stokes[:, p, None]) +
                                            term.sum(axis=1)))
        else:
            raise ValueError("Invalid base %s" % base)

    return spectral_model.reshape(out_shape)


def pol_getter_factory(npoldims):
    if npoldims == 0:
        def impl(pol_shape):
            return 1
    else:
        def impl(pol_shape):
            npols = 1

            for c in pol_shape:
                npols *= c

            return npols

    return njit(nogil=True, cache=True)(impl)


def promote_base_factory(is_base_list):
    if is_base_list:
        def impl(base, npol):
            return base + [base[-1]] * (npol - len(base))
    else:
        def impl(base, npol):
            return [base] * npol

    return njit(nogil=True, cache=True)(impl)


def add_pol_dim_factory(have_pol_dim):
    if have_pol_dim:
        def impl(array):
            return array
    else:
        def impl(array):
            return array.reshape(array.shape + (1,))

    return njit(nogil=True, cache=True)(impl)


@generated_jit(nopython=True, nogil=True, cache=True)
def spectral_model(stokes, spi, ref_freq, frequency, base=0):
    arg_dtypes = tuple(np.dtype(a.dtype.name) for a
                       in (stokes, spi, ref_freq, frequency))
    dtype = np.result_type(*arg_dtypes)

    if isinstance(base, types.containers.List):
        is_base_list = True
        base = base.dtype
    else:
        is_base_list = False

    promote_base = promote_base_factory(is_base_list)

    if isinstance(base, types.scalars.Integer):
        def is_std(base):
            return base == 0

        def is_log(base):
            return base == 1

        def is_log10(base):
            return base == 2

    elif isinstance(base, types.misc.UnicodeType):
        def is_std(base):
            return base == "std"

        def is_log(base):
            return base == "log"

        def is_log10(base):
            return base == "log10"
    else:
        raise TypeError("base '%s' should be a string or integer" % base)

    is_std = njit(nogil=True, cache=True)(is_std)
    is_log = njit(nogil=True, cache=True)(is_log)
    is_log10 = njit(nogil=True, cache=True)(is_log10)

    npoldims = stokes.ndim - 1
    pol_get_fn = pol_getter_factory(npoldims)
    add_pol_dim = add_pol_dim_factory(npoldims > 0)

    if spi.ndim - 2 != npoldims:
        raise ValueError("Dimensions on stokes and spi don't agree")

    def impl(stokes, spi, ref_freq, frequency, base=0):
        nsrc = stokes.shape[0]
        nchan = frequency.shape[0]
        nspi = spi.shape[1]
        npol = pol_get_fn(stokes.shape[1:])

        if npol != pol_get_fn(spi.shape[2:]):
            raise ValueError("Correlations on stokes and spi don't agree")

        # Promote base argument to a per-polarisation list
        list_base = promote_base(base, npol)

        # Reshape adding a polarisation dimension if necessary
        estokes = add_pol_dim(stokes)
        espi = add_pol_dim(spi)

        spectral_model = np.empty((nsrc, nchan, npol), dtype=dtype)

        # TODO(sjperkins)
        # Polarisation + associated base on the outer loop
        # The output cache patterns could be improved.
        for p, b in enumerate(list_base[:npol]):
            if is_std(b):
                for s in range(nsrc):
                    rf = ref_freq[s]

                    for f in range(nchan):
                        freq_ratio = (frequency[f] / rf) - 1.0
                        spec_model = estokes[s, p]

                        for si in range(0, nspi):
                            term = espi[s, si, p] * freq_ratio**(si + 1)
                            spec_model += term

                        spectral_model[s, f, p] = spec_model

            elif is_log(b):
                for s in range(nsrc):
                    rf = ref_freq[s]

                    for f in range(nchan):
                        freq_ratio = np.log(frequency[f] / rf)
                        spec_model = np.log(estokes[s, p])

                        for si in range(0, nspi):
                            term = espi[s, si, p] * freq_ratio**(si + 1)
                            spec_model += term

                        spectral_model[s, f, p] = np.exp(spec_model)

            elif is_log10(b):
                for s in range(nsrc):
                    rf = ref_freq[s]

                    for f in range(nchan):
                        freq_ratio = np.log10(frequency[f] / rf)
                        spec_model = np.log10(estokes[s, p])

                        for si in range(0, nspi):
                            term = espi[s, si, p] * freq_ratio**(si + 1)
                            spec_model += term

                        spectral_model[s, f, p] = 10**spec_model

            else:
                raise ValueError("Invalid base")

        out_shape = (stokes.shape[0], frequency.shape[0]) + stokes.shape[1:]
        return spectral_model.reshape(out_shape)

    return impl


SPECTRAL_MODEL_DOC = DocstringTemplate(r"""
Compute a spectral model, per polarisation.

.. math::
    :nowrap:

    \begin{eqnarray}
    I(\lambda) & = & \sum_{i=0} \alpha_{i} (\lambda / \lambda_0 - 1)^i
      \, \textrm{where} \, \alpha_0 = I(\lambda_0) \\
    \ln( I(\lambda) ) & = & \sum_{i=0} \alpha_{i}
      \ln (\lambda / \lambda_0)^i
      \, \textrm{where} \, \alpha_0 = \ln I_0 \\
    \log_{10}( I(\lambda) ) & = & \sum_{i=0} \alpha_{i}
      \log_{10} (\lambda / \lambda_0)^i
      \, \textrm{where} \, \alpha_0 = \log_{10} I_0 \\
    \end{eqnarray}

Parameters
----------
stokes : $(array_type)
    Stokes parameters of shape :code:`(source,)` or :code:`(source, pol)`.
    If a ``pol`` dimension is present, then it must also be present on ``spi``.
spi : $(array_type)
    Spectral index of shape :code:`(source, spi-comps)`
    or :code:`(source, spi-comps, pol)`.
ref_freq : $(array_type)
    Reference frequencies of shape :code:`(source,)`
frequencies : $(array_type)
    Frequencies of shape :code:`(chan,)`
base : {"std", "log", "log10"} or {0, 1, 2} or list.
    string or corresponding enumeration specifying the polynomial base.
    Defaults to 0.

    If a list is provided, a polynomial base can be specified for each
    stokes parameter or polarisation in the ``pol`` dimension.

    string specification of the base is only supported in python 3.
    while the corresponding integer enumerations are supported
    on all python versions.

Returns
-------
spectral_model : $(array_type)
    Spectral Model of shape :code:`(source, chan)` or
    :code:`(source, chan, pol)`.
""")

try:
    spectral_model.__doc__ = SPECTRAL_MODEL_DOC.substitute(
                                array_type=":class:`numpy.ndarray`")
except AttributeError:
    pass




import numpy as np


def estimate_cell_size(u, v, wavelength, factor=3.0, ny=None, nx=None):
    r"""
    Estimate the cell size in arcseconds given
    baseline ``u`` and ``v`` coordinates, as well
    as the ``wavelengths``, :math:`\lambda`.

    The cell size is computed as:

    .. math::

        \Delta u = 1.0 / \left( 2 \times \text{ factor }
                                  \times \max (\vert u \vert)
                                  / \min( \lambda) \right)

        \Delta v = 1.0 / \left( 2 \times \text{ factor }
                                  \times \max (\vert v \vert)
                                  / \min( \lambda) \right)


    If ``ny`` and ``nx`` are provided the following checks are performed
    and exceptions are raised on failure:

    .. math::

        \Delta u * \text{ ny } \leq \min (\lambda) / \min (\vert u \vert)

        \Delta v * \text{ nx } \leq \min (\lambda) / \min (\vert v \vert)

    Parameters
    ----------
    u : :class:`numpy.ndarray` or float
        Maximum ``u`` coordinate in metres.
    v : :class:`numpy.ndarray` or float
        Maximum ``v`` coordinate in metres.
    wavelength : :class:`numpy.ndarray` or float
        Wavelengths, in metres.
    factor : float, optional
        Scaling factor
    ny : int, optional
        Grid y dimension
    nx : int, optional
        Grid x dimension

    Raises
    ------
    ValueError
        If the cell size criteria are not matched.

    Returns
    -------
    :class:`numpy.ndarray`
        Cell size of ``u`` and ``v`` in arcseconds with shape :code:`(2,)`
    """
    if isinstance(u, np.ndarray):
        abs_u = np.abs(u)
        umax = abs_u.max()
        umin = abs_u.min()
    elif isinstance(u, float):
        umax = umin = abs(u)
    else:
        raise TypeError("Invalid u type %s" % type(u))

    if isinstance(v, np.ndarray):
        abs_v = np.abs(v)
        vmax = abs_v.max()
        vmin = abs_v.min()
    elif isinstance(v, float):
        vmax = vmin = abs(v)
    else:
        raise TypeError("Invalid v type %s" % type(v))

    if isinstance(wavelength, np.ndarray):
        wave_min = wavelength.min()
    elif isinstance(wavelength, float):
        wave_min = wavelength
    else:
        raise TypeError("Invalid wavelength type %s" % type(v))

    umax /= wave_min
    vmax /= wave_min
    umin /= wave_min
    vmin /= wave_min

    u_cell_size = 1.0 / (2.0 * factor * umax)
    v_cell_size = 1.0 / (2.0 * factor * vmax)

    if ny is not None and u_cell_size*ny < (1.0 / umin):
        raise ValueError("v_cell_size*ny [%f] < (1.0 / umin) [%f]" %
                         (u_cell_size*ny, 1.0 / umin))

    if nx is not None and v_cell_size*nx < (1.0 / vmin):
        raise ValueError("v_cell_size*nx [%f] < (1.0 / vmin) [%f]" %
                         (v_cell_size*nx, 1.0 / vmin))

    # Convert radians to arcseconds
    return np.rad2deg([u_cell_size, v_cell_size])*(60*60)



#!/usr/bin/env python
# -*- coding: utf-8 -*-


import argparse
from collections import namedtuple
from functools import lru_cache
from operator import getitem
import weakref

import numpy as np

try:
    from astropy.io import fits
    import dask
    import dask.array as da
    from dask.diagnostics import ProgressBar
    import Tigger
    from daskms import xds_from_ms, xds_from_table, xds_to_table
except ImportError as e:
    opt_import_error = e
else:
    opt_import_error = None

from africanus.util.beams import beam_filenames, beam_grids
from africanus.coordinates.dask import radec_to_lm
from africanus.rime.dask import (phase_delay, predict_vis, parallactic_angles,
                                 beam_cube_dde, feed_rotation)
from africanus.model.coherency.dask import convert
from africanus.model.spectral.dask import spectral_model
from africanus.model.shape.dask import gaussian as gaussian_shape
from africanus.util.requirements import requires_optional


_einsum_corr_indices = 'ijkl'


def _brightness_schema(corrs, index):
    if corrs == 4:
        return "sf" + _einsum_corr_indices[index:index + 2], index + 1
    else:
        return "sfi", index


def _phase_delay_schema(corrs, index):
    return "srf", index


def _spi_schema(corrs, index):
    return "s", index


def _gauss_shape_schema(corrs, index):
    return "srf", index


def _bl_jones_output_schema(corrs, index):
    if corrs == 4:
        return "->srfi" + _einsum_corr_indices[index]
    else:
        return "->srfi"


_rime_term_map = {
    'brightness': _brightness_schema,
    'phase_delay': _phase_delay_schema,
    'spi': _spi_schema,
    'gauss_shape': _gauss_shape_schema,
}


def corr_schema(pol):
    """
    Parameters
    ----------
    pol : Dataset

    Returns
    -------
    corr_schema : list of list
        correlation schema from the POLARIZATION table,
        `[[9, 10], [11, 12]]` for example
    """

    # Select the single row out
    corrs = pol.NUM_CORR.data[0]
    corr_types = pol.CORR_TYPE.data[0]

    if corrs == 4:
        return [[corr_types[0], corr_types[1]],
                [corr_types[2], corr_types[3]]]  # (2, 2) shape
    elif corrs == 2:
        return [corr_types[0], corr_types[1]]    # (2, ) shape
    elif corrs == 1:
        return [corr_types[0]]                   # (1, ) shape
    else:
        raise ValueError("corrs %d not in (1, 2, 4)" % corrs)


def baseline_jones_multiply(corrs, *args):
    names = args[::2]
    arrays = args[1::2]

    input_einsum_schemas = []
    corr_index = 0

    for name, array in zip(names, arrays):
        try:
            # Obtain function for prescribing the input einsum schema
            schema_fn = _rime_term_map[name]
        except KeyError:
            raise ValueError("Unknown RIME term '%s'" % name)
        else:
            # Extract it and the next corr index
            einsum_schema, corr_index = schema_fn(corrs, corr_index)
            input_einsum_schemas.append(einsum_schema)

            if not len(einsum_schema) == array.ndim:
                raise ValueError("%s len(%s) == %d != %s.ndim"
                                 % (name, einsum_schema,
                                    len(einsum_schema), array.shape))

    output_schema = _bl_jones_output_schema(corrs, corr_index)
    schema = ",".join(input_einsum_schemas) + output_schema

    return da.einsum(schema, *arrays)


def create_parser():
    p = argparse.ArgumentParser()
    p.add_argument("ms")
    p.add_argument("-sm", "--sky-model", default="sky-model.txt")
    p.add_argument("-rc", "--row-chunks", type=int, default=10000)
    p.add_argument("-mc", "--model-chunks", type=int, default=10)
    p.add_argument("-b", "--beam", default=None)
    p.add_argument("-iuvw", "--invert-uvw", action="store_true",
                   help="Invert UVW coordinates. Useful if we want "
                        "compare our visibilities against MeqTrees")
    return p


@lru_cache(maxsize=16)
def load_beams(beam_file_schema, corr_types):

    class FITSFile(object):
        """ Exists so that fits file is closed when last ref is gc'd """

        def __init__(self, filename):
            self.hdul = hdul = fits.open(filename)
            assert len(hdul) == 1
            self.__del_ref = weakref.ref(self, lambda r: hdul.close())

    # Open files and get headers
    beam_files = []
    headers = []

    for corr, (re, im) in beam_filenames(beam_file_schema, corr_types).items():
        re_f = FITSFile(re)
        im_f = FITSFile(im)
        beam_files.append((corr, (re_f, im_f)))
        headers.append((corr, (re_f.hdul[0].header, im_f.hdul[0].header)))

    # All FITS headers should agree
    flat_headers = [d for k, v in headers for d in v]

    if not all(flat_headers[0] == h for h in flat_headers[1:]):
        raise ValueError("BEAM FITS Header Files differ")

    #  Map FITS header type to NumPy type
    BITPIX_MAP = {8: np.dtype('uint8').type, 16: np.dtype('int16').type,
                  32: np.dtype('int32').type, -32: np.dtype('float32').type,
                  -64: np.dtype('float64').type}

    header = flat_headers[0]
    bitpix = header['BITPIX']

    try:
        dtype = BITPIX_MAP[bitpix]
    except KeyError:
        raise ValueError("No mapping from BITPIX %s to a numpy type" % bitpix)
    else:
        dtype = np.result_type(dtype, np.complex64)

    if not header['NAXIS'] == 3:
        raise ValueError("FITS must have exactly three axes. "
                         "L or X, M or Y and FREQ. NAXIS != 3")

    (l_ax, l_grid), (m_ax, m_grid), (nu_ax, nu_grid) = beam_grids(header)

    # Shape of each correlation
    shape = (l_grid.shape[0], m_grid.shape[0], nu_grid.shape[0])

    # Axis tranpose, FITS is FORTRAN ordered
    ax = (nu_ax - 1, m_ax - 1, l_ax - 1)

    def _load_correlation(re, im, ax):
        # Read real and imaginary for each correlation
        return (re.hdul[0].data.transpose(ax) +
                im.hdul[0].data.transpose(ax)*1j)

    # Create delayed loads of the beam
    beam_loader = dask.delayed(_load_correlation)

    beam_corrs = [beam_loader(re, im, ax)
                  for c, (corr, (re, im)) in enumerate(beam_files)]
    beam_corrs = [da.from_delayed(bc, shape=shape, dtype=dtype)
                  for bc in beam_corrs]

    # Stack correlations and rechunk to one great big block
    beam = da.stack(beam_corrs, axis=3)
    beam = beam.rechunk(shape + (len(corr_types),))

    # Dask arrays for the beam extents and beam frequency grid
    beam_lm_ext = np.array([[l_grid[0], l_grid[-1]], [m_grid[0], m_grid[-1]]])
    beam_lm_ext = da.from_array(beam_lm_ext, chunks=beam_lm_ext.shape)
    beam_freq_grid = da.from_array(nu_grid, chunks=nu_grid.shape)

    return beam, beam_lm_ext, beam_freq_grid


def parse_sky_model(filename, chunks):
    """
    Parses a Tigger sky model

    Parameters
    ----------
    filename : str
        Sky Model filename
    chunks : tuple of ints or int
        Source chunking strategy

    Returns
    -------
    source_data : dict
        Dictionary of source data,
        :code:`{'point': (...), 'gauss': (...) }`
    """
    sky_model = Tigger.load(filename, verbose=False)

    _empty_spectrum = object()

    point_radec = []
    point_stokes = []
    point_spi = []
    point_ref_freq = []

    gauss_radec = []
    gauss_stokes = []
    gauss_spi = []
    gauss_ref_freq = []
    gauss_shape = []

    for source in sky_model.sources:
        ra = source.pos.ra
        dec = source.pos.dec
        typecode = source.typecode.lower()

        I = source.flux.I  # noqa
        Q = source.flux.Q
        U = source.flux.U
        V = source.flux.V

        spectrum = (getattr(source, "spectrum", _empty_spectrum)
                    or _empty_spectrum)

        try:
            # Extract reference frequency
            ref_freq = spectrum.freq0
        except AttributeError:
            ref_freq = sky_model.freq0

        try:
            # Extract SPI for I.
            # Zero Q, U and V to get 1 on the exponential
            spi = [[spectrum.spi, 0, 0, 0]]
        except AttributeError:
            # Default I SPI to -0.7
            spi = [[-0.7, 0, 0, 0]]

        if typecode == "gau":
            emaj = source.shape.ex
            emin = source.shape.ey
            pa = source.shape.pa

            gauss_radec.append([ra, dec])
            gauss_stokes.append([I, Q, U, V])
            gauss_spi.append(spi)
            gauss_ref_freq.append(ref_freq)
            gauss_shape.append([emaj, emin, pa])

        elif typecode == "pnt":
            point_radec.append([ra, dec])
            point_stokes.append([I, Q, U, V])
            point_spi.append(spi)
            point_ref_freq.append(ref_freq)
        else:
            raise ValueError("Unknown source morphology %s" % typecode)

    Point = namedtuple("Point", ["radec", "stokes", "spi", "ref_freq"])
    Gauss = namedtuple("Gauss", ["radec", "stokes", "spi", "ref_freq",
                                 "shape"])

    source_data = {}

    if len(point_radec) > 0:
        source_data['point'] = Point(
                    da.from_array(point_radec, chunks=(chunks, -1)),
                    da.from_array(point_stokes, chunks=(chunks, -1)),
                    da.from_array(point_spi, chunks=(chunks, 1, -1)),
                    da.from_array(point_ref_freq, chunks=chunks))

    if len(gauss_radec) > 0:
        source_data['gauss'] = Gauss(
                    da.from_array(gauss_radec, chunks=(chunks, -1)),
                    da.from_array(gauss_stokes, chunks=(chunks, -1)),
                    da.from_array(gauss_spi, chunks=(chunks, 1, -1)),
                    da.from_array(gauss_ref_freq, chunks=chunks),
                    da.from_array(gauss_shape, chunks=(chunks, -1)))

    return source_data


def support_tables(args):
    """
    Parameters
    ----------
    args : object
        Script argument objects

    Returns
    -------
    table_map : dict of Dataset
        {name: dataset}
    """

    n = {k: '::'.join((args.ms, k)) for k
         in ("ANTENNA", "DATA_DESCRIPTION", "FIELD",
             "SPECTRAL_WINDOW", "POLARIZATION")}

    # All rows at once
    lazy_tables = {"ANTENNA": xds_from_table(n["ANTENNA"])}

    compute_tables = {
        # Fixed shape rows
        "DATA_DESCRIPTION": xds_from_table(n["DATA_DESCRIPTION"]),
        # Variably shaped, need a dataset per row
        "FIELD": xds_from_table(n["FIELD"],
                                group_cols="__row__"),
        "SPECTRAL_WINDOW": xds_from_table(n["SPECTRAL_WINDOW"],
                                          group_cols="__row__"),
        "POLARIZATION": xds_from_table(n["POLARIZATION"],
                                       group_cols="__row__"),
    }

    lazy_tables.update(dask.compute(compute_tables)[0])
    return lazy_tables


def _zero_pes(parangles, frequency, dtype_):
    """ Create zeroed pointing errors """
    ntime, na = parangles.shape
    nchan = frequency.shape[0]
    return np.zeros((ntime, na, nchan, 2), dtype=dtype_)


def _unity_ant_scales(parangles, frequency, dtype_):
    """ Create zeroed antenna scalings """
    _, na = parangles[0].shape
    nchan = frequency.shape[0]
    return np.ones((na, nchan, 2), dtype=dtype_)


def dde_factory(args, ms, ant, field, pol, lm, utime, frequency):
    if args.beam is None:
        return None

    # Beam is requested
    corr_type = tuple(pol.CORR_TYPE.data[0])

    if not len(corr_type) == 4:
        raise ValueError("Need four correlations for DDEs")

    parangles = parallactic_angles(utime, ant.POSITION.data,
                                   field.PHASE_DIR.data[0][0])

    corr_type_set = set(corr_type)

    if corr_type_set.issubset(set([9, 10, 11, 12])):
        pol_type = 'linear'
    elif corr_type_set.issubset(set([5, 6, 7, 8])):
        pol_type = 'circular'
    else:
        raise ValueError("Cannot determine polarisation type "
                         "from correlations %s. Constructing "
                         "a feed rotation matrix will not be "
                         "possible." % (corr_type,))

    # Construct feed rotation
    feed_rot = feed_rotation(parangles, pol_type)

    dtype = np.result_type(parangles, frequency)

    # Create zeroed pointing errors
    zpe = da.blockwise(_zero_pes, ("time", "ant", "chan", "comp"),
                       parangles, ("time", "ant"),
                       frequency, ("chan",),
                       dtype, None,
                       new_axes={"comp": 2},
                       dtype=dtype)

    # Created zeroed antenna scaling factors
    zas = da.blockwise(_unity_ant_scales, ("ant", "chan", "comp"),
                       parangles, ("time", "ant"),
                       frequency, ("chan",),
                       dtype, None,
                       new_axes={"comp": 2},
                       dtype=dtype)

    # Load the beam information
    beam, lm_ext, freq_map = load_beams(args.beam, corr_type)

    # Introduce the correlation axis
    beam = beam.reshape(beam.shape[:3] + (2, 2))

    beam_dde = beam_cube_dde(beam, lm_ext, freq_map, lm, parangles,
                             zpe, zas,
                             frequency)

    # Multiply the beam by the feed rotation to form the DDE term
    return da.einsum("stafij,tajk->stafik", beam_dde, feed_rot)


def vis_factory(args, source_type, sky_model,
                ms, ant, field, spw, pol):
    try:
        source = sky_model[source_type]
    except KeyError:
        raise ValueError("Source type '%s' unsupported" % source_type)

    # Select single dataset rows
    corrs = pol.NUM_CORR.data[0]
    frequency = spw.CHAN_FREQ.data[0]
    phase_dir = field.PHASE_DIR.data[0][0]  # row, poly

    lm = radec_to_lm(source.radec, phase_dir)
    uvw = -ms.UVW.data if args.invert_uvw else ms.UVW.data

    # (source, row, frequency)
    phase = phase_delay(lm, uvw, frequency)

    # (source, spi, corrs)
    # Apply spectral mode to stokes parameters
    stokes = spectral_model(source.stokes,
                            source.spi,
                            source.ref_freq,
                            frequency,
                            base=[1, 0, 0, 0])

    brightness = convert(stokes, ["I", "Q", "U", "V"],
                         corr_schema(pol))

    bl_jones_args = ["phase_delay", phase]

    # Add any visibility amplitude terms
    if source_type == "gauss":
        bl_jones_args.append("gauss_shape")
        bl_jones_args.append(gaussian_shape(uvw, frequency, source.shape))

    bl_jones_args.extend(["brightness", brightness])

    # Unique times and time index for each row chunk
    # The index is not global
    meta = np.empty((0,), dtype=tuple)
    utime_inv = ms.TIME.data.map_blocks(np.unique, return_inverse=True,
                                        meta=meta, dtype=tuple)

    # Need unique times for parallactic angles
    utime = utime_inv.map_blocks(getitem, 0,
                                 chunks=(np.nan,),
                                 dtype=ms.TIME.dtype)

    time_idx = utime_inv.map_blocks(getitem, 1, dtype=np.int32)

    jones = baseline_jones_multiply(corrs, *bl_jones_args)
    dde = dde_factory(args, ms, ant, field, pol, lm, utime, frequency)

    return predict_vis(time_idx, ms.ANTENNA1.data, ms.ANTENNA2.data,
                       dde, jones, dde, None, None, None)


@requires_optional("dask.array", "Tigger",
                   "daskms", opt_import_error)
def predict(args):
    # Convert source data into dask arrays
    sky_model = parse_sky_model(args.sky_model, args.model_chunks)

    # Get the support tables
    tables = support_tables(args)

    ant_ds = tables["ANTENNA"]
    field_ds = tables["FIELD"]
    ddid_ds = tables["DATA_DESCRIPTION"]
    spw_ds = tables["SPECTRAL_WINDOW"]
    pol_ds = tables["POLARIZATION"]

    # List of write operations
    writes = []

    # Construct a graph for each DATA_DESC_ID
    for xds in xds_from_ms(args.ms,
                           columns=["UVW", "ANTENNA1", "ANTENNA2", "TIME"],
                           group_cols=["FIELD_ID", "DATA_DESC_ID"],
                           chunks={"row": args.row_chunks}):

        # Perform subtable joins
        ant = ant_ds[0]
        field = field_ds[xds.attrs['FIELD_ID']]
        ddid = ddid_ds[xds.attrs['DATA_DESC_ID']]
        spw = spw_ds[ddid.SPECTRAL_WINDOW_ID.data[0]]
        pol = pol_ds[ddid.POLARIZATION_ID.data[0]]

        # Select single dataset row out
        corrs = pol.NUM_CORR.data[0]

        # Generate visibility expressions for each source type
        source_vis = [vis_factory(args, stype, sky_model,
                                  xds, ant, field, spw, pol)
                      for stype in sky_model.keys()]

        # Sum visibilities together
        vis = sum(source_vis)

        # Reshape (2, 2) correlation to shape (4,)
        if corrs == 4:
            vis = vis.reshape(vis.shape[:2] + (4,))

        # Assign visibilities to MODEL_DATA array on the dataset
        xds = xds.assign(MODEL_DATA=(("row", "chan", "corr"), vis))
        # Create a write to the table
        write = xds_to_table(xds, args.ms, ['MODEL_DATA'])
        # Add to the list of writes
        writes.append(write)

    # Submit all graph computations in parallel
    with ProgressBar():
        dask.compute(writes)


if __name__ == "__main__":
    args = create_parser().parse_args()
    predict(args)



# -*- coding: utf-8 -*-


from collections import namedtuple

import numba
import numpy as np

from africanus.constants import minus_two_pi_over_c
from africanus.util.docs import doc_tuple_to_str
from africanus.util.numba import is_numba_type_none, generated_jit


@generated_jit(nopython=True, nogil=True, cache=True)
def im_to_vis(image, uvw, lm, frequency, dtype=None):
    # Infer complex output dtype if none provided
    if is_numba_type_none(dtype):
        out_dtype = np.result_type(np.complex64,
                                   *(np.dtype(a.dtype.name) for a in
                                     (image, uvw, lm, frequency)))
    else:
        out_dtype = dtype.dtype

    def _im_to_vis_impl(image, uvw, lm, frequency, dtype=None):
        nrows = uvw.shape[0]
        nsrc = lm.shape[0]
        nchan = frequency.shape[0]
        ncorr = image.shape[-1]
        vis_of_im = np.zeros((nrows, nchan, ncorr), dtype=out_dtype)

        # For each uvw coordinate
        for r in range(nrows):
            u, v, w = uvw[r]

            # For each source
            for s in range(nsrc):
                l, m = lm[s]
                n = np.sqrt(1.0 - l**2 - m**2) - 1.0

                # e^(-2*pi*(l*u + m*v + n*w)/c)
                real_phase = minus_two_pi_over_c * (l * u + m * v + n * w)

                # Multiple in frequency for each channel
                for nu in range(nchan):
                    p = real_phase * frequency[nu] * 1.0j

                    for c in range(ncorr):
                        if image[s, nu, c]:
                            vis_of_im[r, nu, c] += np.exp(p)*image[s, nu, c]

        return vis_of_im

    return _im_to_vis_impl


@generated_jit(nopython=True, nogil=True, cache=True)
def vis_to_im(vis, uvw, lm, frequency, flags, dtype=None):
    # Infer output dtype if none provided
    if is_numba_type_none(dtype):
        # Support both real and complex visibilities...
        if isinstance(vis.dtype, numba.types.scalars.Complex):
            vis_comp_dtype = np.dtype(vis.dtype.underlying_float.name)
        else:
            vis_comp_dtype = np.dtype(vis.dtype.name)

        out_dtype = np.result_type(vis_comp_dtype,
                                   *(np.dtype(a.dtype.name) for a in
                                     (uvw, lm, frequency)))
    else:
        if isinstance(dtype, numba.types.scalars.Complex):
            raise TypeError("dtype must be complex")

        out_dtype = dtype.dtype

    assert np.shape(vis) == np.shape(flags)

    def _vis_to_im_impl(vis, uvw, lm, frequency, flags, dtype=None):
        nrows = uvw.shape[0]
        nsrc = lm.shape[0]
        nchan = frequency.shape[0]
        ncorr = vis.shape[-1]

        im_of_vis = np.zeros((nsrc, nchan, ncorr), dtype=out_dtype)

        # For each source
        for s in range(nsrc):
            l, m = lm[s]
            n = np.sqrt(1.0 - l ** 2 - m ** 2) - 1.0
            # For each uvw coordinate
            for r in range(nrows):
                u, v, w = uvw[r]

                # e^(-2*pi*(l*u + m*v + n*w)/c)
                real_phase = -minus_two_pi_over_c * (l * u + m * v + n * w)

                # Multiple in frequency for each channel
                for nu in range(nchan):
                    p = real_phase * frequency[nu]

                    # do not compute if any of the correlations
                    # are flagged (complicates uncertainties)
                    if np.any(flags[r, nu]):
                        continue

                    for c in range(ncorr):
                        # elide the call to exp since result is real
                        im_of_vis[s, nu, c] += (np.cos(p) *
                                                vis[r, nu, c].real -
                                                np.sin(p) *
                                                vis[r, nu, c].imag)

        return im_of_vis

    return _vis_to_im_impl


_DFT_DOCSTRING = namedtuple(
    "_DFTDOCSTRING", ["preamble", "parameters", "returns"])

im_to_vis_docs = _DFT_DOCSTRING(
    preamble="""
    Computes the discrete image to visibility mapping
    of an ideal interferometer:

    .. math::

        {\\Large \\sum_s e^{-2 \\pi i (u l_s + v m_s + w (n_s - 1))} \\cdot I_s }

    """,  # noqa

    parameters="""
    Parameters
    ----------

    image : :class:`numpy.ndarray`
        image of shape :code:`(source, chan, corr)`
        The brighness matrix in each pixel (flatten 2D array
        per channel and corr). Note not Stokes terms
    uvw : :class:`numpy.ndarray`
        uvw coordinates of shape :code:`(row, 3)` with
        u, v and w components in the last dimension.
    lm : :class:`numpy.ndarray`
        lm coordinates of shape :code:`(source, 2)` with
        l and m components in the last dimension.
    frequency : :class:`numpy.ndarray`
        frequencies of shape :code:`(chan,)`
    dtype : np.dtype, optional
        Datatype of result. Should be either np.complex64 or np.complex128.
        If ``None``, :func:`numpy.result_type` is used to infer the data type
        from the inputs.
    """,

    returns="""
    Returns
    -------
    visibilties : :class:`numpy.ndarray`
        complex of shape :code:`(row, chan, corr)`
    """
)


im_to_vis.__doc__ = doc_tuple_to_str(im_to_vis_docs)

vis_to_im_docs = _DFT_DOCSTRING(
    preamble="""
    Computes visibility to image mapping
    of an ideal interferometer:

    .. math::

        {\\Large \\sum_k e^{ 2 \\pi i (u_k l + v_k m + w_k (n - 1))} \\cdot V_k}

    """,  # noqa

    parameters="""
    Parameters
    ----------

    vis : :class:`numpy.ndarray`
        visibilities of shape :code:`(row, chan, corr)`
        Visibilities corresponding to brightness terms.
        Note the dirty images produced do not necessarily
        correspond to Stokes terms and need to be converted.
    uvw : :class:`numpy.ndarray`
        uvw coordinates of shape :code:`(row, 3)` with
        u, v and w components in the last dimension.
    lm : :class:`numpy.ndarray`
        lm coordinates of shape :code:`(source, 2)` with
        l and m components in the last dimension.
    frequency : :class:`numpy.ndarray`
        frequencies of shape :code:`(chan,)`
    flags : :class:`numpy.ndarray`
        Boolean array of shape :code:`(row, chan, corr)`
        Note that if one correlation is flagged we discard
        all of them otherwise we end up irretrievably
        mixing Stokes terms.
    dtype : np.dtype, optional
        Datatype of result. Should be either np.float32 or np.float64.
        If ``None``, :func:`numpy.result_type` is used to infer the data type
        from the inputs.
    """,

    returns="""
    Returns
    -------
    image : :class:`numpy.ndarray`
        float of shape :code:`(source, chan, corr)`
    """
)


vis_to_im.__doc__ = doc_tuple_to_str(vis_to_im_docs)
